# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2016 Rackspace Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import shlex
import socket
import subprocess
import tempfile
import time

from oslo_log import log as logging
from six.moves import http_cookiejar
from six.moves.urllib import error
from six.moves.urllib import request as urllib2
from tempest.common import utils
from tempest.common import waiters
from tempest import config
from tempest.lib.common.utils import test_utils
from tempest.lib import exceptions as lib_exc

from neutron_lbaas._i18n import _
from neutron_lbaas.tests.tempest.v2.clients import health_monitors_client
from neutron_lbaas.tests.tempest.v2.clients import listeners_client
from neutron_lbaas.tests.tempest.v2.clients import load_balancers_client
from neutron_lbaas.tests.tempest.v2.clients import members_client
from neutron_lbaas.tests.tempest.v2.clients import pools_client
from neutron_lbaas.tests.tempest.v2.scenario import manager

config = config.CONF

LOG = logging.getLogger(__name__)


def _setup_config_args(auth_provider):
    """Set up ServiceClient arguments using config settings. """
    service = config.network.catalog_type
    region = config.network.region or config.identity.region
    endpoint_type = config.network.endpoint_type
    build_interval = config.network.build_interval
    build_timeout = config.network.build_timeout

    # The disable_ssl appears in identity
    disable_ssl_certificate_validation = (
        config.identity.disable_ssl_certificate_validation)
    ca_certs = None

    # Trace in debug section
    trace_requests = config.debug.trace_requests

    return [auth_provider, service, region, endpoint_type, build_interval,
            build_timeout, disable_ssl_certificate_validation, ca_certs,
            trace_requests]


class BaseTestCase(manager.NetworkScenarioTest):

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.servers_keypairs = {}
        self.servers = {}
        self.members = []
        self.floating_ips = {}
        self.servers_floating_ips = {}
        self.server_ips = {}
        self.port1 = 80
        self.port2 = 88
        self.num = 50
        self.server_fixed_ips = {}

        self.listener_protocol = config.lbaas.default_listener_protocol
        self.pool_protocol = config.lbaas.default_pool_protocol
        self.hm_protocol = config.lbaas.default_health_monitor_protocol

        self._create_security_group_for_test()
        self._set_net_and_subnet()

        mgr = self.get_client_manager()
        auth_provider = mgr.auth_provider
        self.client_args = _setup_config_args(auth_provider)

        self.load_balancers_client = (
            load_balancers_client.LoadBalancersClientJSON(*self.client_args))
        self.listeners_client = (
            listeners_client.ListenersClientJSON(*self.client_args))
        self.pools_client = pools_client.PoolsClientJSON(*self.client_args)
        self.members_client = members_client.MembersClientJSON(
            *self.client_args)
        self.health_monitors_client = (
            health_monitors_client.HealthMonitorsClientJSON(
                *self.client_args))

    @classmethod
    def skip_checks(cls):
        super(BaseTestCase, cls).skip_checks()
        cfg = config.network
        if not utils.is_extension_enabled('lbaasv2', 'network'):
            msg = 'LBaaS Extension is not enabled'
            raise cls.skipException(msg)
        if not (cfg.project_networks_reachable or cfg.public_network_id):
            msg = ('Either tenant_networks_reachable must be "true", or '
                   'public_network_id must be defined.')
            raise cls.skipException(msg)

    def _set_net_and_subnet(self):
        """
        Query and set appropriate network and subnet attributes to be used
        for the test. Existing tenant networks are used if they are found.
        The configured private network and associated subnet is used as a
        fallback in absence of tenant networking.
        """
        tenant_id = self.networks_client.tenant_id
        try:
            tenant_net = self.os_admin.networks_client.list_networks(
                tenant_id=tenant_id)['networks'][0]
        except IndexError:
            tenant_net = None

        if tenant_net:
            self.subnet = self.os_admin.subnets_client.list_subnets(
                tenant_id=tenant_id)['subnets'][0]
            self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                            self.networks_client.delete_network,
                            self.subnet['id'])
            self.network = tenant_net
        else:
            self.network = self._get_network_by_name(
                config.compute.fixed_network_name)
            # We are assuming that the first subnet associated
            # with the fixed network is the one we want.  In the future, we
            # should instead pull a subnet id from config, which is set by
            # devstack/admin/etc.
            subnet = self.os_admin.subnets_client.list_subnets(
                network_id=self.network['id'])['subnets'][0]
            self.subnet = subnet

    def _create_security_group_for_test(self):
        self.security_group = self._create_security_group(
            tenant_id=self.networks_client.tenant_id)
        self._create_security_group_rules_for_port(self.port1)
        self._create_security_group_rules_for_port(self.port2)

    def _create_security_group_rules_for_port(self, port):
        rule = {
            'direction': 'ingress',
            'protocol': 'tcp',
            'port_range_min': port,
            'port_range_max': port,
        }
        self._create_security_group_rule(
            secgroup=self.security_group,
            tenant_id=self.networks_client.tenant_id,
            **rule)

    def _ipv6_subnet(self, address6_mode):
        tenant_id = self.networks_client.tenant_id
        router = self._get_router(tenant_id=tenant_id)
        self.network = self._create_network(tenant_id=tenant_id)
        self.subnet = self._create_subnet(network=self.network,
                                          namestart='sub6',
                                          ip_version=6,
                                          ipv6_ra_mode=address6_mode,
                                          ipv6_address_mode=address6_mode)
        self.subnet.add_to_router(router_id=router['id'])
        self.addCleanup(self.subnet.delete)

    def _create_server(self, name):
        keypair = self.create_keypair()
        security_groups = [{'name': self.security_group['name']}]
        create_kwargs = {
            'networks': [
                {'uuid': self.network['id']},
            ],
            'key_name': keypair['name'],
            'security_groups': security_groups,
            'name': name
        }
        net_name = self.network['name']
        server = self.create_server(**create_kwargs)
        waiters.wait_for_server_status(self.servers_client,
                                       server['id'], 'ACTIVE')
        server = self.servers_client.show_server(server['id'])
        server = server['server']
        self.servers_keypairs[server['id']] = keypair
        if (config.network.public_network_id and not
                config.network.project_networks_reachable):
            public_network_id = config.network.public_network_id
            floating_ip = self.create_floating_ip(
                server, public_network_id)
            self.floating_ips[floating_ip['id']] = server
            self.server_ips[server['id']] = floating_ip['floating_ip_address']
        else:
            self.server_ips[server['id']] =\
                server['addresses'][net_name][0]['addr']
        self.server_fixed_ips[server['id']] =\
            server['addresses'][net_name][0]['addr']
        self.assertTrue(self.servers_keypairs)
        return server

    def _create_servers(self):
        for count in range(2):
            self.server = self._create_server(name=("server%s" % (count + 1)))
            if count == 0:
                self.servers['primary'] = self.server['id']
            else:
                self.servers['secondary'] = self.server['id']
        self.assertEqual(len(self.servers_keypairs), 2)

    def _stop_server(self):
        for name, value in self.servers.items():
            if name == 'primary':
                self.servers_client.stop_server(value)
                waiters.wait_for_server_status(self.servers_client,
                                               value, 'SHUTOFF')

    def _start_server(self):
        for name, value in self.servers.items():
            if name == 'primary':
                self.servers_client.start(value)
                waiters.wait_for_server_status(self.servers_client,
                                               value, 'ACTIVE')

    def _start_servers(self):
        """
        Start two backends
        1. SSH to the instance
        2. Start two http backends listening on ports 80 and 88 respectively
        """
        for server_id, ip in self.server_ips.items():
            private_key = self.servers_keypairs[server_id]['private_key']
            server = self.servers_client.show_server(server_id)['server']
            server_name = server['name']
            username = config.validation.image_ssh_user
            ssh_client = self.get_remote_client(
                ip_address=ip,
                private_key=private_key)

            # Write a backend's response into a file
            resp = ('#!/bin/sh\n'
                    'echo -ne "HTTP/1.1 200 OK\r\nContent-Length: 7\r\n'
                    'Set-Cookie:JSESSIONID=%(s_id)s\r\nConnection: close\r\n'
                    'Content-Type: text/html; '
                    'charset=UTF-8\r\n\r\n%(server)s"; cat >/dev/null')

            with tempfile.NamedTemporaryFile(mode='w+') as script:
                script.write(resp % {'s_id': server_name[-1],
                                     'server': server_name})
                script.flush()
                with tempfile.NamedTemporaryFile(mode='w+') as key:
                    key.write(private_key)
                    key.flush()
                    self.copy_file_to_host(script.name,
                                           "/tmp/script1",
                                           ip,
                                           username, key.name)

            # Start netcat
            start_server = ('chmod a+x /tmp/%(script)s; '
                            'while true; do '
                            'sudo nc -ll -p %(port)s -e /tmp/%(script)s; '
                            'done > /dev/null &')
            cmd = start_server % {'port': self.port1,
                                  'script': 'script1'}
            ssh_client.exec_command(cmd)

            if len(self.server_ips) == 1:
                with tempfile.NamedTemporaryFile(mode='w+') as script:
                    script.write(resp % {'s_id': 2,
                                         'server': 'server2'})
                    script.flush()
                    with tempfile.NamedTemporaryFile(mode='w+') as key:
                        key.write(private_key)
                        key.flush()
                        self.copy_file_to_host(script.name,
                                               "/tmp/script2", ip,
                                               username, key.name)
                cmd = start_server % {'port': self.port2,
                                      'script': 'script2'}
                ssh_client.exec_command(cmd)

    def _create_listener(self, load_balancer_id, port=80, **kwargs):
        """Create a listener with HTTP protocol listening on port 80."""
        self.listener = self.listeners_client.create_listener(
            loadbalancer_id=load_balancer_id, protocol=self.listener_protocol,
            protocol_port=port, **kwargs)
        self.assertTrue(self.listener)
        self.addCleanup(self._cleanup_listener, self.listener.get('id'))
        return self.listener

    def _create_health_monitor(self):
        """Create a pool with ROUND_ROBIN algorithm."""
        self.hm = self.health_monitors_client.create_health_monitor(
            type=self.hm_protocol, max_retries=5, delay=3, timeout=5,
            pool_id=self.pool['id'])
        self.assertTrue(self.hm)
        self.addCleanup(self._cleanup_health_monitor, self.hm.get('id'))

    def _create_pool(self, listener_id, persistence_type=None,
                     cookie_name=None):
        """Create a pool with ROUND_ROBIN algorithm."""
        pool = {
            "listener_id": listener_id,
            "lb_algorithm": "ROUND_ROBIN",
            "protocol": self.pool_protocol
        }
        if persistence_type:
            pool.update({'session_persistence': {'type': persistence_type}})
        if cookie_name:
            pool.update({'session_persistence': {"cookie_name": cookie_name}})
        self.pool = self.pools_client.create_pool(**pool)
        self.assertTrue(self.pool)
        self.addCleanup(self._cleanup_pool, self.pool['id'])
        return self.pool

    def _cleanup_load_balancer(self, load_balancer_id):
        test_utils.call_and_ignore_notfound_exc(
            self.load_balancers_client.delete_load_balancer, load_balancer_id)
        self._wait_for_load_balancer_deletion(load_balancer_id)

    def _cleanup_listener(self, listener_id):
        test_utils.call_and_ignore_notfound_exc(
            self.listeners_client.delete_listener, listener_id)
        self._wait_for_listener_deletion(listener_id)

    def _cleanup_pool(self, pool_id):
        test_utils.call_and_ignore_notfound_exc(
            self.pools_client.delete_pool, pool_id)
        self._wait_for_pool_deletion(pool_id)

    def _cleanup_health_monitor(self, hm_id):
        test_utils.call_and_ignore_notfound_exc(
            self.health_monitors_client.delete_health_monitor, hm_id)
        self._wait_for_health_monitor_deletion(hm_id)

    def _create_members(self, load_balancer_id=None, pool_id=None,
                        subnet_id=None):
        """
        Create two members.

        In case there is only one server, create both members with the same ip
        but with different ports to listen on.
        """
        for server_id, ip in self.server_fixed_ips.items():
            if len(self.server_fixed_ips) == 1:
                member1 = self.members_client.create_member(
                    pool_id=pool_id,
                    address=ip,
                    protocol_port=self.port1,
                    subnet_id=subnet_id)
                self._wait_for_load_balancer_status(load_balancer_id)
                member2 = self.members_client.create_member(
                    pool_id=pool_id,
                    address=ip,
                    protocol_port=self.port2,
                    subnet_id=subnet_id)
                self._wait_for_load_balancer_status(load_balancer_id)
                self.members.extend([member1, member2])
            else:
                member = self.members_client.create_member(
                    pool_id=pool_id,
                    address=ip,
                    protocol_port=self.port1,
                    subnet_id=subnet_id)
                self._wait_for_load_balancer_status(load_balancer_id)
                self.members.append(member)
        self.assertTrue(self.members)

    def _assign_floating_ip_to_lb_vip(self, lb):
        public_network_id = config.network.public_network_id
        port_id = lb['vip_port_id']
        floating_ip = self.create_floating_ip(lb, public_network_id,
                                              port_id=port_id)
        self.floating_ips.setdefault(lb['id'], [])
        self.floating_ips[lb['id']].append(floating_ip)
        # Check for floating ip status before you check load-balancer
        self.check_floating_ip_status(floating_ip, "ACTIVE")

    def _create_load_balancer(self, ip_version=4, persistence_type=None):
        tenant_id = self.networks_client.tenant_id
        self.create_lb_kwargs = {'tenant_id': tenant_id,
                                 'vip_subnet_id': self.subnet['id']}
        self.load_balancer = self.load_balancers_client.create_load_balancer(
            **self.create_lb_kwargs)
        load_balancer_id = self.load_balancer['id']
        self.addCleanup(self._cleanup_load_balancer, load_balancer_id)
        self._wait_for_load_balancer_status(load_balancer_id)

        listener = self._create_listener(load_balancer_id=load_balancer_id)
        self._wait_for_load_balancer_status(load_balancer_id)

        self.pool = self._create_pool(listener_id=listener.get('id'),
                                      persistence_type=persistence_type)
        self._wait_for_load_balancer_status(load_balancer_id)

        self._create_members(load_balancer_id=load_balancer_id,
                             pool_id=self.pool['id'],
                             subnet_id=self.subnet['id'])

        self.vip_ip = self.load_balancer.get('vip_address')

        # if the ipv4 is used for lb, then fetch the right values from
        # tempest.conf file
        if ip_version == 4:
            if (config.network.public_network_id and not
                    config.network.project_networks_reachable):
                self._assign_floating_ip_to_lb_vip(self.load_balancer)
                self.vip_ip = self.floating_ips[
                    self.load_balancer['id']][0]['floating_ip_address']

        # Currently the ovs-agent is not enforcing security groups on the
        # vip port - see https://bugs.launchpad.net/neutron/+bug/1163569
        # However the linuxbridge-agent does, and it is necessary to add a
        # security group with a rule that allows tcp port 80 to the vip port.
        self.ports_client.update_port(
            self.load_balancer.get('vip_port_id'),
            security_groups=[self.security_group['id']])

    def _wait_for_load_balancer_status(self, load_balancer_id,
                                       provisioning_status='ACTIVE',
                                       operating_status='ONLINE'):
        interval_time = 1
        timeout = 600
        end_time = time.time() + timeout
        while time.time() < end_time:
            try:
                lb = self.load_balancers_client.get_load_balancer(
                    load_balancer_id)
            except lib_exc.NotFound:
                    raise
            if (lb.get('provisioning_status') == provisioning_status and
                    lb.get('operating_status') == operating_status):
                break
            elif (lb.get('provisioning_status') == 'ERROR' or
                    lb.get('operating_status') == 'ERROR'):
                raise Exception(
                    _("Wait for load balancer for load balancer: {lb_id} "
                      "ran for {timeout} seconds and an ERROR was encountered "
                      "with provisioning status: {provisioning_status} and "
                      "operating status: {operating_status}").format(
                          timeout=timeout,
                          lb_id=lb.get('id'),
                          provisioning_status=provisioning_status,
                          operating_status=operating_status))
            time.sleep(interval_time)
        else:
            raise Exception(
                _("Wait for load balancer ran for {timeout} seconds and did "
                  "not observe {lb_id} reach {provisioning_status} "
                  "provisioning status and {operating_status} "
                  "operating status.").format(
                      timeout=timeout,
                      lb_id=lb.get('id'),
                      provisioning_status=provisioning_status,
                      operating_status=operating_status))
        return lb

    def _wait_for_resource_deletion(self, resource_type_name,
                                    resource_get_method,
                                    resource_id):
        interval_time = 1
        timeout = 600
        end_time = time.time() + timeout
        while time.time() < end_time:
            try:
                resource_get_method(resource_id)
            except lib_exc.NotFound:
                return
            time.sleep(interval_time)
        else:
            raise Exception(
                _("Wait for {res_name} ran for {timeout} seconds and did "
                  "not observe {res_id} deletion processes ended").format(
                      res_name=resource_type_name,
                      timeout=timeout,
                      res_id=resource_id))

    def _wait_for_load_balancer_deletion(self, load_balancer_id):
        return self._wait_for_resource_deletion(
            'load balancer',
            self.load_balancers_client.get_load_balancer,
            load_balancer_id)

    def _wait_for_pool_deletion(self, pool_id):
        return self._wait_for_resource_deletion(
            'pool',
            self.pools_client.get_pool,
            pool_id)

    def _wait_for_listener_deletion(self, listener_id):
        return self._wait_for_resource_deletion(
            'listener',
            self.listeners_client.get_listener,
            listener_id)

    def _wait_for_health_monitor_deletion(self, health_monitor_id):
        return self._wait_for_resource_deletion(
            'health monitor',
            self.health_monitors_client.get_health_monitor,
            health_monitor_id)

    def _wait_for_pool_session_persistence(self, pool_id, sp_type=None):
        interval_time = 1
        timeout = 10
        end_time = time.time() + timeout
        while time.time() < end_time:
            pool = self.pools_client.get_pool(pool_id)
            sp = pool.get('session_persistence', None)
            if (not (sp_type or sp) or
                    pool['session_persistence']['type'] == sp_type):
                return pool
            time.sleep(interval_time)
        raise Exception(
            _("Wait for pool ran for {timeout} seconds and did "
              "not observe {pool_id} update session persistence type "
              "to {type}.").format(
                  timeout=timeout,
                  pool_id=pool_id,
                  type=sp_type))

    def _check_load_balancing(self, port=80):
        """
        1. Send NUM requests on the floating ip associated with the VIP
        2. Check that the requests are shared between the two servers
        """

        self._check_connection(self.vip_ip, port=port)
        counters = self._send_requests(self.vip_ip, ["server1", "server2"])
        for member, counter in counters.items():
            self.assertGreater(counter, 0, 'Member %s never balanced' % member)

    def _check_connection(self, check_ip, port=80):
        def try_connect(check_ip, port):
            try:
                resp = urllib2.urlopen("http://{0}:{1}/".format(check_ip,
                                                                port))
                if resp.getcode() == 200:
                    return True
                return False
            except IOError:
                return False
            except error.HTTPError:
                return False
        timeout = config.validation.ping_timeout
        start = time.time()
        while not try_connect(check_ip, port):
            if (time.time() - start) > timeout:
                message = ("Timed out trying to connect to {0}:{1} after "
                           "{2} seconds".format(check_ip, port, timeout))
                raise lib_exc.TimeoutException(message)

    def _send_requests(self, vip_ip, servers):
        counters = dict.fromkeys(servers, 0)
        for i in range(self.num):
            try:
                server = urllib2.urlopen("http://{0}/".format(vip_ip),
                                         None, 2).read().decode('utf8')
                counters[server] += 1
            # HTTP exception means fail of server, so don't increase counter
            # of success and continue connection tries
            except (error.HTTPError, error.URLError,
                    socket.timeout, socket.error):
                continue
        return counters

    def _traffic_validation_after_stopping_server(self):
        """Check that the requests are sent to the only ACTIVE server."""
        counters = self._send_requests(self.vip_ip, ["server1", "server2"])

        # Assert that no traffic is sent to server1.
        for member, counter in counters.items():
            if member == 'server1':
                self.assertEqual(counter, 0,
                                 'Member %s is not balanced' % member)

    def _check_load_balancing_after_deleting_resources(self):
        """
        Check that the requests are not sent to any servers
        Assert that no traffic is sent to any servers
        """
        counters = self._send_requests(self.vip_ip, ["server1", "server2"])
        for member, counter in counters.items():
            self.assertEqual(counter, 0, 'Member %s is balanced' % member)

    def _check_source_ip_persistence(self):
        """Check source ip session persistence.

        Verify that all requests from our ip are answered by the same server
        that handled it the first time.
        """
        # Check that backends are reachable
        self._check_connection(self.vip_ip)

        resp = []
        for count in range(10):
            resp.append(
                urllib2.urlopen("http://{0}/".format(self.vip_ip)).read())
        self.assertEqual(len(set(resp)), 1)

    def _update_pool_session_persistence(self, persistence_type=None,
                                         cookie_name=None):
        """Update a pool with new session persistence type and cookie name."""

        update_data = {'session_persistence': None}
        if persistence_type:
            update_data = {"session_persistence": {
                "type": persistence_type}}
        if cookie_name:
            update_data['session_persistence'].update(
                {"cookie_name": cookie_name})
        self.pools_client.update_pool(self.pool['id'], **update_data)
        self.pool = self._wait_for_pool_session_persistence(self.pool['id'],
                                                            persistence_type)
        self._wait_for_load_balancer_status(self.load_balancer['id'])
        if persistence_type:
            self.assertEqual(persistence_type,
                             self.pool['session_persistence']['type'])
        if cookie_name:
            self.assertEqual(cookie_name,
                             self.pool['session_persistence']['cookie_name'])

    def _check_cookie_session_persistence(self):
        """Check cookie persistence types by injecting cookies in requests."""

        # Send first request and get cookie from the server's response
        cj = http_cookiejar.CookieJar()
        opener = urllib2.build_opener(urllib2.HTTPCookieProcessor(cj))
        opener.open("http://{0}/".format(self.vip_ip))
        resp = []
        # Send 10 subsequent requests with the cookie inserted in the headers.
        for count in range(10):
            request = urllib2.Request("http://{0}/".format(self.vip_ip))
            cj.add_cookie_header(request)
            response = urllib2.urlopen(request)
            resp.append(response.read())
        self.assertEqual(len(set(resp)), 1, message=resp)

    def copy_file_to_host(self, file_from, dest, host, username, pkey):
        dest = "%s@%s:%s" % (username, host, dest)
        cmd = ("scp -v -o UserKnownHostsFile=/dev/null "
               "-o StrictHostKeyChecking=no "
               "-i %(pkey)s %(file1)s %(dest)s" % {'pkey': pkey,
                                                   'file1': file_from,
                                                   'dest': dest})
        args = shlex.split(cmd)
        subprocess_args = {'stdout': subprocess.PIPE,
                           'stderr': subprocess.STDOUT}
        proc = subprocess.Popen(args, **subprocess_args)
        stdout, stderr = proc.communicate()
        if proc.returncode != 0:
            LOG.error(("Command {0} returned with exit status {1},"
                      "output {2}, error {3}").format(cmd, proc.returncode,
                                                      stdout, stderr))
        return stdout



# Copyright 2015, 2016 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import utils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as ex

from neutron_lbaas.tests.tempest.v2.api import base


class TestHealthMonitors(base.BaseTestCase):

    """
    Tests the following operations in the Neutron-LBaaS API using the
    REST client for Health Monitors:
        list health monitors
        create health monitor
        get health monitor
        update health monitor
        delete health monitor
    """

    @classmethod
    def resource_setup(cls):
        super(TestHealthMonitors, cls).resource_setup()
        if not utils.is_extension_enabled('lbaasv2', 'network'):
            msg = "lbaas extension not enabled."
            raise cls.skipException(msg)
        network_name = data_utils.rand_name('network-')
        cls.network = cls.create_network(network_name)
        cls.subnet = cls.create_subnet(cls.network)
        cls.load_balancer = cls._create_load_balancer(
            tenant_id=cls.subnet.get('tenant_id'),
            vip_subnet_id=cls.subnet.get('id'))
        cls.listener = cls._create_listener(
            loadbalancer_id=cls.load_balancer.get('id'),
            protocol=cls.listener_protocol, protocol_port=80)
        cls.pool = cls._create_pool(
            protocol=cls.pool_protocol, lb_algorithm='ROUND_ROBIN',
            listener_id=cls.listener.get('id'))
        cls.create_basic_hm_kwargs = {'type': cls.hm_protocol, 'delay': 3,
                                      'max_retries': 10, 'timeout': 5,
                                      'pool_id': cls.pool.get('id')}

    def _prep_list_comparison(self, single, obj_list):
        single.pop('operating_status', None)
        for obj in obj_list:
            obj.pop('operating_status', None)
            if not single.get('updated_at') and obj.get('updated_at'):
                obj['updated_at'] = None
            self._test_provisioning_status_if_exists(single, obj)

    def test_list_health_monitors_empty(self):
        hm_list = self.health_monitors_client.list_health_monitors()
        self.assertEmpty(hm_list)

    def test_list_health_monitors_one(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        hm_list = self.health_monitors_client.list_health_monitors()
        self._prep_list_comparison(hm, hm_list)
        self.assertIn(hm, hm_list)

    @decorators.attr(type='smoke')
    def test_list_health_monitors_two(self):
        hm1 = self._create_health_monitor(**self.create_basic_hm_kwargs)
        new_listener = self._create_listener(
            loadbalancer_id=self.load_balancer.get('id'),
            protocol=self.listener_protocol, protocol_port=88)
        self.addCleanup(self._delete_listener, new_listener.get('id'))
        new_pool = self._create_pool(
            protocol=self.pool_protocol, lb_algorithm='ROUND_ROBIN',
            listener_id=new_listener.get('id'))
        self.addCleanup(self._delete_pool, new_pool.get('id'))
        hm2 = self._create_health_monitor(
            type='HTTPS',
            max_retries=3,
            delay=1,
            timeout=2,
            pool_id=new_pool.get('id'))
        hm_list = self.health_monitors_client.list_health_monitors()
        self._prep_list_comparison(hm1, hm_list)
        self._prep_list_comparison(hm2, hm_list)
        self.assertEqual(2, len(hm_list))
        self.assertIn(hm1, hm_list)
        self.assertIn(hm2, hm_list)

    @decorators.attr(type='smoke')
    def test_create_and_get_health_monitor(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        hm_test = self.health_monitors_client.get_health_monitor(hm.get('id'))
        self._test_provisioning_status_if_exists(hm, hm_test)
        hm.pop('operating_status', None)
        hm_test.pop('operating_status', None)
        self.assertEqual(hm, hm_test)

    def test_create_health_monitor_missing_attribute(self):
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          pool_id=self.pool.get('id'))

    def test_create_health_monitor_missing_required_field_type(self):
        """Test if a non_admin user can create a health monitor with type
        missing
        """
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          delay=3, max_retries=10, timeout=5,
                          pool_id=self.pool.get('id'))

    def test_create_health_monitor_missing_required_field_delay(self):
        """Test if a non_admin user can create a health monitor with delay
        missing
        """
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, max_retries=10, timeout=5,
                          pool_id=self.pool.get('id'))

    def test_create_health_monitor_missing_required_field_timeout(self):
        """Test if a non_admin user can create a health monitor with timeout
        missing
        """
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          pool_id=self.pool.get('id'))

    def test_create_health_monitor_missing_required_field_max_retries(self):
        """Test if a non_admin user can create a health monitor with max_retries
        missing
        """
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, timeout=5,
                          pool_id=self.pool.get('id'))

    def test_create_health_monitor_missing_required_field_pool_id(self):
        """Test if a non_admin user can create a health monitor with pool_id
        missing
        """
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout=5)

    def test_create_health_monitor_missing_admin_state_up(self):
        """Test if a non_admin user can create a health monitor with
        admin_state_up missing
        """
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        hm_test = self.health_monitors_client.get_health_monitor(hm.get('id'))
        self._test_provisioning_status_if_exists(hm, hm_test)
        hm.pop('operating_status', None)
        hm_test.pop('operating_status', None)
        self.assertEqual(hm, hm_test)
        self.assertTrue(hm_test.get('admin_state_up'))

    def test_create_health_monitor_missing_http_method(self):
        """Test if a non_admin user can create a health monitor with
        http_method missing
        """
        if self.hm_protocol != 'HTTP':
            msg = "health monitor protocol must be HTTP for http_method"
            raise self.skipException(msg)
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)

        hm_test = self.health_monitors_client.get_health_monitor(hm.get('id'))
        self._test_provisioning_status_if_exists(hm, hm_test)
        hm.pop('operating_status', None)
        hm_test.pop('operating_status', None)
        self.assertEqual(hm, hm_test)
        self.assertEqual('GET', hm_test.get('http_method'))

    def test_create_health_monitor_missing_url_path(self):
        """Test if a non_admin user can create a health monitor with
        url_path missing
        """
        if self.hm_protocol != 'HTTP':
            msg = "health monitor protocol must be HTTP for url_path"
            raise self.skipException(msg)
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        hm_test = self.health_monitors_client.get_health_monitor(hm.get('id'))
        self._test_provisioning_status_if_exists(hm, hm_test)
        hm.pop('operating_status', None)
        hm_test.pop('operating_status', None)
        self.assertEqual(hm, hm_test)
        self.assertEqual('/', hm_test.get('url_path'))

    def test_create_health_monitor_missing_expected_codes(self):
        """Test if a non_admin user can create a health monitor with
        expected_codes missing
        """
        if self.hm_protocol != 'HTTP':
            msg = "health monitor protocol must be HTTP for expected_codes"
            raise self.skipException(msg)
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)

        hm_test = self.health_monitors_client.get_health_monitor(hm.get('id'))
        self._test_provisioning_status_if_exists(hm, hm_test)
        hm.pop('operating_status', None)
        hm_test.pop('operating_status', None)
        self.assertEqual(hm, hm_test)
        self.assertEqual('200', hm_test.get('expected_codes'))

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_create_health_monitor_invalid_tenant_id(self):
        """Test create health monitor with invalid tenant_id"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          tenant_id='blah',
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout=5, pool_id=self.pool.get('id'))

    @decorators.attr(type='negative')
    def test_create_health_monitor_invalid_type(self):
        """Test create health monitor with invalid type"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type='blah', delay=3, max_retries=10, timeout=5,
                          pool_id=self.pool.get('id'))

    @decorators.attr(type='negative')
    def test_create_health_monitor_invalid_delay(self):
        """Test create health monitor with invalid delay"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay='blah', max_retries=10,
                          timeout=5, pool_id=self.pool.get('id'))

    @decorators.attr(type='negative')
    def test_create_health_monitor_invalid_max_retries(self):
        """Test create health monitor with invalid max_retries"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries='blah',
                          timeout=5, pool_id=self.pool.get('id'))

    @decorators.attr(type='negative')
    def test_create_health_monitor_invalid_timeout(self):
        """Test create health monitor with invalid timeout"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout='blah', pool_id=self.pool.get('id'))

    @decorators.attr(type='negative')
    def test_create_health_monitor_invalid_pool_id(self):
        """Test create health monitor with invalid pool id"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout=5, pool_id='blah')

    @decorators.attr(type='negative')
    def test_create_health_monitor_invalid_admin_state_up(self):
        """Test if a non_admin user can create a health monitor with invalid
        admin_state_up
        """
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout=5, pool_id=self.pool.get('id'),
                          admin_state_up='blah')

    @decorators.attr(type='negative')
    def test_create_health_monitor_invalid_expected_codes(self):
        """Test if a non_admin user can create a health monitor with invalid
        expected_codes
        """
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout=5, pool_id=self.pool.get('id'),
                          expected_codes='blah')

    @decorators.attr(type='negative')
    def test_create_health_monitor_invalid_url_path(self):
        """Test if a non_admin user can create a health monitor with invalid
        url_path
        """
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout=5, pool_id=self.pool.get('id'),
                          url_path='blah')

    @decorators.attr(type='negative')
    def test_create_health_monitor_invalid_http_method(self):
        """Test if a non_admin user can create a health monitor with invalid
        http_method
        """
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout=5, pool_id=self.pool.get('id'),
                          http_method='blah')

    @decorators.attr(type='negative')
    def test_create_health_monitor_empty_type(self):
        """Test create health monitor with empty type"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type='', delay=3, max_retries=10, timeout=5,
                          pool_id=self.pool.get('id'))

    @decorators.attr(type='negative')
    def test_create_health_monitor_empty_delay(self):
        """Test create health monitor with empty delay"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay='', max_retries=10,
                          timeout=5, pool_id=self.pool.get('id'))

    @decorators.attr(type='negative')
    def test_create_health_monitor_empty_timeout(self):
        """Test create health monitor with empty timeout"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout='', pool_id=self.pool.get('id'))

    @decorators.attr(type='negative')
    def test_create_health_monitor_empty_max_retries(self):
        """Test create health monitor with empty max_retries"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries='',
                          timeout=5, pool_id=self.pool.get('id'))

    @decorators.attr(type='negative')
    def test_create_health_monitor_empty_max_pool_id(self):
        """Test create health monitor with empty pool_id"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout=5, pool_id='')

    @decorators.attr(type='negative')
    def test_create_health_monitor_empty_max_admin_state_up(self):
        """Test create health monitor with empty admin_state_up"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout=5, pool_id=self.pool.get('id'),
                          admin_state_up='')

    @decorators.attr(type='negative')
    def test_create_health_monitor_empty_max_http_method(self):
        """Test create health monitor with empty http_method"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout=5, pool_id=self.pool.get('id'),
                          http_method='')

    @decorators.attr(type='negative')
    def test_create_health_monitor_empty_max_url_path(self):
        """Test create health monitor with empty url_path"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout=5, pool_id=self.pool.get('id'), url_path='')

    @decorators.attr(type='negative')
    def test_create_health_monitor_empty_expected_codes(self):
        """Test create health monitor with empty expected_codes"""
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout=5, pool_id=self.pool.get('id'),
                          expected_codes='')

    @decorators.attr(type='negative')
    def test_create_health_monitor_invalid_attribute(self):
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3,
                          max_retries='twenty one',
                          pool_id=self.pool.get('id'))

    @decorators.attr(type='negative')
    def test_create_health_monitor_extra_attribute(self):
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          pool_id=self.pool.get('id'), subnet_id=10)

    @decorators.attr(type='smoke')
    def test_update_health_monitor(self):
        hm = self._create_health_monitor(type=self.hm_protocol, delay=3,
                                         max_retries=10, timeout=5,
                                         pool_id=self.pool.get('id'))
        max_retries = 1
        self._update_health_monitor(hm.get('id'), max_retries=max_retries)
        new_hm = self.health_monitors_client.get_health_monitor(hm.get('id'))
        self.assertEqual(max_retries, new_hm.get('max_retries'))

    def test_update_health_monitor_missing_admin_state_up(self):
        """Test update health monitor with missing admin state field"""
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self._update_health_monitor(hm.get('id'))
        new_hm = self.health_monitors_client.get_health_monitor(hm.get('id'))
        self.assertTrue(new_hm.get('admin_state_up'))

    def test_update_health_monitor_missing_delay(self):
        """Test update health monitor with missing delay field"""
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self._update_health_monitor(hm.get('id'))
        new_hm = self.health_monitors_client.get_health_monitor(hm.get('id'))
        self.assertEqual(hm.get('delay'), new_hm.get('delay'))

    def test_update_health_monitor_missing_timeout(self):
        """Test update health monitor with missing timeout field"""
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self._update_health_monitor(hm.get('id'))
        new_hm = self.health_monitors_client.get_health_monitor(hm.get('id'))
        self.assertEqual(hm.get('timeout'), new_hm.get('timeout'))

    def test_update_health_monitor_missing_max_retries(self):
        """Test update health monitor with missing max retries field"""
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self._update_health_monitor(hm.get('id'))
        new_hm = self.health_monitors_client.get_health_monitor(hm.get('id'))
        self.assertEqual(hm.get('max_retries'), new_hm.get('max_retries'))

    def test_update_health_monitor_missing_http_method(self):
        """Test update health monitor with missing http_method field"""
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self._update_health_monitor(hm.get('id'))
        new_hm = self.health_monitors_client.get_health_monitor(hm.get('id'))
        self.assertEqual(hm.get('http_method'), new_hm.get('http_method'))

    def test_update_health_monitor_missing_url_path(self):
        """Test update health monitor with missing url_path field"""
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self._update_health_monitor(hm.get('id'))
        new_hm = self.health_monitors_client.get_health_monitor(hm.get('id'))
        self.assertEqual(hm.get('url_path'), new_hm.get('url_path'))

    def test_update_health_monitor_missing_expected_codes(self):
        """Test update health monitor with missing expected_codes field"""
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self._update_health_monitor(hm.get('id'))
        new_hm = self.health_monitors_client.get_health_monitor(hm.get('id'))
        self.assertEqual(hm.get('expected_codes'),
                         new_hm.get('expected_codes'))

    @decorators.attr(type='negative')
    def test_update_health_monitor_invalid_attribute(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), max_retries='blue')

    @decorators.attr(type='negative')
    def test_update_health_monitor_invalid_admin_state_up(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), admin_state_up='blah')

    @decorators.attr(type='negative')
    def test_update_health_monitor_invalid_delay(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), delay='blah')

    @decorators.attr(type='negative')
    def test_update_health_monitor_invalid_timeout(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), timeout='blah')

    @decorators.attr(type='negative')
    def test_update_health_monitor_invalid_max_retries(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), max_retries='blah')

    @decorators.attr(type='negative')
    def test_update_health_monitor_invalid_http_method(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), http_method='blah')

    @decorators.attr(type='negative')
    def test_update_health_monitor_invalid_url_path(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), url_path='blah')

    @decorators.attr(type='negative')
    def test_update_health_monitor_invalid_expected_codes(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), expected_codes='blah')

    @decorators.attr(type='negative')
    def test_update_health_monitor_empty_admin_state_up(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), admin_state_up='')

    @decorators.attr(type='negative')
    def test_update_health_monitor_empty_delay(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), empty_delay='')

    @decorators.attr(type='negative')
    def test_update_health_monitor_empty_timeout(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), timeout='')

    @decorators.attr(type='negative')
    def test_update_health_monitor_empty_max_retries(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), max_retries='')

    @decorators.attr(type='negative')
    def test_update_health_monitor_empty_empty_http_method(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), http_method='')

    @decorators.attr(type='negative')
    def test_update_health_monitor_empty_url_path(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), http_method='')

    @decorators.attr(type='negative')
    def test_update_health_monitor_empty_expected_codes(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), expected_codes='')

    def test_update_health_monitor_extra_attribute(self):
        hm = self._create_health_monitor(**self.create_basic_hm_kwargs)
        self.assertRaises(ex.BadRequest,
                          self._update_health_monitor,
                          hm.get('id'), protocol='UDP')

    @decorators.attr(type='smoke')
    def test_delete_health_monitor(self):
        hm = self._create_health_monitor(cleanup=False, type=self.hm_protocol,
                                         delay=3, max_retries=10, timeout=5,
                                         pool_id=self.pool.get('id'))
        self._delete_health_monitor(hm.get('id'))
        self.assertRaises(ex.NotFound,
                          self.health_monitors_client.get_health_monitor,
                          hm.get('id'))



# Copyright 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import extensions
from neutron_lib import constants as n_constants
from neutron_lib import exceptions as nexception

from neutron.api.v2 import resource_helper
from neutron_lib.plugins import constants

from neutron_lbaas._i18n import _
from neutron_lbaas.extensions import loadbalancerv2


class ProviderCannotCreateLoadBalancerGraph(nexception.BadRequest):
    message = _("The provider does not have the ability to create a load "
                "balancer graph.")

# NOTE(blogan): this dictionary is to be used only for importing from the
# plugin to validate against.  It is only put here for consistency with
# all other extensions and an easy place to look what changes this extension
# allows.
RESOURCE_ATTRIBUTE_MAP = {
    'graphs': {
        'loadbalancer': {'allow_post': True, 'allow_put': False,
                         'is_visible': True},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'is_visible': True}
    }
}

EXISTING_ATTR_GRAPH_ATTR_MAP = {
    'loadbalancers': {
        'listeners': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True, 'default': []
        }
    },
    'listeners': {
        'default_pool': {
            'allow_post': True, 'allow_put': False, 'is_visible': True,
            'default': n_constants.ATTR_NOT_SPECIFIED
        },
        'l7policies': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True, 'default': []
        }
    },
    'pools': {
        'healthmonitor': {
            'allow_post': True, 'allow_put': False, 'is_visible': True,
            'default': n_constants.ATTR_NOT_SPECIFIED
        },
        'members': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True, 'default': []
        }
    },
    'l7policies': {
        'rules': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True, 'default': []
        },
        'redirect_pool': {
            'allow_post': True, 'allow_put': False, 'is_visible': True,
            'default': n_constants.ATTR_NOT_SPECIFIED
        },
        'listener_id': {
            'allow_post': False, 'allow_put': False, 'is_visible': True
        }
    }
}


class Lb_graph(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Load Balancer Graph"

    @classmethod
    def get_alias(cls):
        return "lb-graph"

    @classmethod
    def get_description(cls):
        return "Extension for allowing the creation of load balancers with a" \
               " full graph in one API request."

    @classmethod
    def get_updated(cls):
        return "2016-02-09T10:00:00-00:00"

    def get_required_extensions(self):
        return ["lbaasv2"]

    @classmethod
    def get_resources(cls):
        plural_mappings = resource_helper.build_plural_mappings(
            {}, RESOURCE_ATTRIBUTE_MAP)
        resources = resource_helper.build_resource_info(
            plural_mappings,
            RESOURCE_ATTRIBUTE_MAP,
            constants.LOADBALANCERV2,
            register_quota=False)
        return resources

    @classmethod
    def get_plugin_interface(cls):
        return loadbalancerv2.LoadBalancerPluginBaseV2



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from neutron_lbaas.drivers import driver_base

LOG = logging.getLogger(__name__)


class LoggingNoopLoadBalancerDriver(driver_base.LoadBalancerBaseDriver):

    def __init__(self, plugin):
        super(LoggingNoopLoadBalancerDriver, self).__init__(plugin)

        # Each of the major LBaaS objects in the neutron database
        # need a corresponding manager/handler class.
        #
        # Put common things that are shared across the entire driver, like
        # config or a rest client handle, here.
        #
        # This function is executed when neutron-server starts.

        self.load_balancer = LoggingNoopLoadBalancerManager(self)
        self.listener = LoggingNoopListenerManager(self)
        self.pool = LoggingNoopPoolManager(self)
        self.member = LoggingNoopMemberManager(self)
        self.health_monitor = LoggingNoopHealthMonitorManager(self)
        self.l7policy = LoggingNoopL7PolicyManager(self)
        self.l7rule = LoggingNoopL7RuleManager(self)


class LoggingNoopCommonManager(object):

    @driver_base.driver_op
    def create(self, context, obj):
        LOG.debug("LB %s no-op, create %s", self.__class__.__name__, obj.id)

    @driver_base.driver_op
    def update(self, context, old_obj, obj):
        LOG.debug("LB %s no-op, update %s", self.__class__.__name__, obj.id)

    @driver_base.driver_op
    def delete(self, context, obj):
        LOG.debug("LB %s no-op, delete %s", self.__class__.__name__, obj.id)


class LoggingNoopLoadBalancerManager(LoggingNoopCommonManager,
                                     driver_base.BaseLoadBalancerManager):

    @property
    def allows_create_graph(self):
        return True

    @property
    def allows_healthmonitor_thresholds(self):
        return True

    @property
    def allocates_vip(self):
        LOG.debug('allocates_vip queried')
        return False

    def create_and_allocate_vip(self, context, obj):
        LOG.debug("LB %s no-op, create_and_allocate_vip %s",
                  self.__class__.__name__, obj.id)
        self.create(context, obj)

    @driver_base.driver_op
    def refresh(self, context, obj):
        # This is intended to trigger the backend to check and repair
        # the state of this load balancer and all of its dependent objects
        LOG.debug("LB pool refresh %s", obj.id)

    @driver_base.driver_op
    def stats(self, context, lb_obj):
        LOG.debug("LB stats %s", lb_obj.id)
        return {
            "bytes_in": 0,
            "bytes_out": 0,
            "active_connections": 0,
            "total_connections": 0
        }


class LoggingNoopListenerManager(LoggingNoopCommonManager,
                                 driver_base.BaseListenerManager):
    pass


class LoggingNoopPoolManager(LoggingNoopCommonManager,
                             driver_base.BasePoolManager):
    pass


class LoggingNoopMemberManager(LoggingNoopCommonManager,
                               driver_base.BaseMemberManager):
    pass


class LoggingNoopHealthMonitorManager(LoggingNoopCommonManager,
                                      driver_base.BaseHealthMonitorManager):
    pass


class LoggingNoopL7PolicyManager(LoggingNoopCommonManager,
                                 driver_base.BaseL7PolicyManager):
    pass


class LoggingNoopL7RuleManager(LoggingNoopCommonManager,
                               driver_base.BaseL7RuleManager):
    pass



#    Copyright 2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1.identity import v2 as v2_client
from keystoneauth1.identity import v3 as v3_client
from keystoneauth1 import session
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from neutron_lbaas._i18n import _


LOG = logging.getLogger(__name__)

_SESSION = None
OPTS = [
    cfg.StrOpt(
        'auth_url',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        help=_('Authentication endpoint'),
    ),
    cfg.StrOpt(
        'admin_user',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        default='admin',
        help=_('The service admin user name'),
    ),
    cfg.StrOpt(
        'admin_tenant_name',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        default='admin',
        help=_('The service admin tenant name'),
    ),
    cfg.StrOpt(
        'admin_password',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        secret=True,
        default='password',
        help=_('The service admin password'),
    ),
    cfg.StrOpt(
        'admin_user_domain',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        default='Default',
        help=_('The admin user domain name'),
    ),
    cfg.StrOpt(
        'admin_project_domain',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        default='Default',
        help=_('The admin project domain name'),
    ),
    cfg.StrOpt(
        'region',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        default='RegionOne',
        help=_('The deployment region'),
    ),
    cfg.StrOpt(
        'service_name',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        default='lbaas',
        help=_('The name of the service'),
    ),
    cfg.StrOpt(
        'auth_version',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        default='3',
        help=_('The auth version used to authenticate'),
    ),
    cfg.StrOpt(
        'endpoint_type',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        default='public',
        help=_('The endpoint_type to be used')
    ),
    cfg.BoolOpt(
        'insecure',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        default=False,
        help=_('Disable server certificate verification')
    ),
    cfg.StrOpt(
        'cafile',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        help=_('CA certificate file path')
    ),
    cfg.StrOpt(
        'certfile',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        help=_('Client certificate cert file path')
    ),
    cfg.StrOpt(
        'keyfile',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        help=_('Client certificate key file path')
    )
]

cfg.CONF.register_opts(OPTS, 'service_auth')


def get_session():
    """Initializes a Keystone session.

    :returns: a Keystone Session object
    :raises Exception: if the session cannot be established
    """
    global _SESSION
    if not _SESSION:

        auth_url = cfg.CONF.service_auth.auth_url
        insecure = cfg.CONF.service_auth.insecure
        cacert = cfg.CONF.service_auth.cafile
        cert = cfg.CONF.service_auth.certfile
        key = cfg.CONF.service_auth.keyfile

        if insecure:
            verify = False
        else:
            verify = cacert or True

        if cert and key:
            cert = (cert, key)

        kwargs = {'auth_url': auth_url,
                  'username': cfg.CONF.service_auth.admin_user,
                  'password': cfg.CONF.service_auth.admin_password}

        if cfg.CONF.service_auth.auth_version == '2':
            client = v2_client
            kwargs['tenant_name'] = cfg.CONF.service_auth.admin_tenant_name
        elif cfg.CONF.service_auth.auth_version == '3':
            client = v3_client
            kwargs['project_name'] = cfg.CONF.service_auth.admin_tenant_name
            kwargs['user_domain_name'] = (cfg.CONF.service_auth.
                                          admin_user_domain)
            kwargs['project_domain_name'] = (cfg.CONF.service_auth.
                                             admin_project_domain)
        else:
            raise Exception(_('Unknown keystone version!'))

        try:
            kc = client.Password(**kwargs)
            _SESSION = session.Session(auth=kc, verify=verify, cert=cert)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("Error creating Keystone session.")

    return _SESSION



# Copyright 2016 Rackspace
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib.api import extensions

EXTENDED_ATTRIBUTES_2_0 = {
    'healthmonitors': {
        'max_retries_down': {
            'allow_post': True, 'allow_put': True,
            'default': 3, 'validate': {'type:range': [1, 10]},
            'convert_to': converters.convert_to_int, 'is_visible': True
        }
    }
}


class Healthmonitor_max_retries_down(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Add a fall threshold to health monitor"

    @classmethod
    def get_alias(cls):
        return "hm_max_retries_down"

    @classmethod
    def get_description(cls):
        return "Add a fall threshold to health monitor"

    @classmethod
    def get_updated(cls):
        return "2016-04-19T16:00:00-00:00"

    def get_required_extensions(self):
        return ["lbaasv2"]

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



# Copyright 2013 New Dream Network, LLC (DreamHost)
# Copyright 2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from neutron.common import config as common_config
from neutron.common import rpc as n_rpc
from neutron.conf.agent import common as config
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import service

from neutron_lbaas._i18n import _
from neutron_lbaas.agent import agent_manager as manager
from neutron_lbaas.services.loadbalancer import constants

LOG = logging.getLogger(__name__)

OPTS = [
    cfg.IntOpt(
        'periodic_interval',
        default=10,
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        help=_('Seconds between periodic task runs')
    )
]


class LbaasAgentService(n_rpc.Service):
    def start(self):
        super(LbaasAgentService, self).start()
        self.tg.add_timer(
            cfg.CONF.periodic_interval,
            self.manager.run_periodic_tasks,
            None,
            None
        )


def main():
    cfg.CONF.register_opts(OPTS)
    cfg.CONF.register_opts(manager.OPTS)
    # import interface options just in case the driver uses namespaces
    config.register_interface_opts(cfg.CONF)
    config.register_external_process_opts(cfg.CONF)
    config.register_interface_driver_opts_helper(cfg.CONF)
    config.register_agent_state_opts_helper(cfg.CONF)
    config.register_root_helper(cfg.CONF)

    common_config.init(sys.argv[1:])
    config.setup_logging()
    config.setup_privsep()

    LOG.warning('neutron-lbaas is now deprecated. See: '
                'https://wiki.openstack.org/wiki/Neutron/LBaaS/Deprecation')

    mgr = manager.LbaasAgentManager(cfg.CONF)
    svc = LbaasAgentService(
        host=cfg.CONF.host,
        topic=constants.LOADBALANCER_AGENTV2,
        manager=mgr
    )
    service.launch(cfg.CONF, svc).wait()



#
# Copyright 2014 Brocade Communications Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Pattabi Ayyasami (pattabi), Brocade Communications Systems,Inc.
#

from brocade_neutron_lbaas import adx_device_driver_v2 as device_driver

from neutron_lbaas.drivers import driver_base


class BrocadeLoadBalancerDriver(driver_base.LoadBalancerBaseDriver):

    def __init__(self, plugin):
        super(BrocadeLoadBalancerDriver, self).__init__(plugin)

        self.load_balancer = BrocadeLoadBalancerManager(self)
        self.listener = BrocadeListenerManager(self)
        self.pool = BrocadePoolManager(self)
        self.member = BrocadeMemberManager(self)
        self.health_monitor = BrocadeHealthMonitorManager(self)
        self.device_driver = device_driver.BrocadeAdxDeviceDriverV2(plugin)


class BrocadeLoadBalancerManager(driver_base.BaseLoadBalancerManager):
    def create(self, context, obj):
        try:
            self.driver.device_driver.create_loadbalancer(obj)
            self.successful_completion(context, obj)
        except Exception:
            self.failed_completion(context, obj)

    def update(self, context, old_obj, obj):
        try:
            self.driver.device_driver.update_loadbalancer(obj, old_obj)
            self.successful_completion(context, obj)
        except Exception:
            self.failed_completion(context, obj)

    def delete(self, context, obj):
        try:
            self.driver.device_driver.delete_loadbalancer(obj)
        except Exception:
            # Ignore the exception
            pass

        self.successful_completion(context, obj, delete=True)

    def refresh(self, context, lb_obj):
        # This is intended to trigger the backend to check and repair
        # the state of this load balancer and all of its dependent objects
        self.driver.device_driver.refresh(lb_obj)

    def stats(self, context, lb_obj):
        return self.driver.device_driver.stats(lb_obj)


class BrocadeListenerManager(driver_base.BaseListenerManager):
    def create(self, context, obj):
        try:
            self.driver.device_driver.create_listener(obj)
            self.successful_completion(context, obj)
        except Exception:
            self.failed_completion(context, obj)

    def update(self, context, old_obj, obj):
        try:
            self.driver.device_driver.update_listener(obj, old_obj)
            self.successful_completion(context, obj)
        except Exception:
            self.failed_completion(context, obj)

    def delete(self, context, obj):
        try:
            self.driver.device_driver.delete_listener(obj)
        except Exception:
            # Ignore the exception
            pass

        self.successful_completion(context, obj, delete=True)


class BrocadePoolManager(driver_base.BasePoolManager):
    def create(self, context, obj):
        try:
            self.driver.device_driver.create_pool(obj)
            self.successful_completion(context, obj)
        except Exception:
            self.failed_completion(context, obj)

    def update(self, context, old_obj, obj):
        try:
            self.driver.device_driver.update_pool(obj, old_obj)
            self.successful_completion(context, obj)
        except Exception:
            self.failed_completion(context, obj)

    def delete(self, context, obj):
        try:
            self.driver.device_driver.delete_pool(obj)
        except Exception:
            # Ignore the exception
            pass

        self.successful_completion(context, obj, delete=True)


class BrocadeMemberManager(driver_base.BaseMemberManager):
    def create(self, context, obj):
        try:
            self.driver.device_driver.create_member(obj)
            self.successful_completion(context, obj)
        except Exception:
            self.failed_completion(context, obj)

    def update(self, context, old_obj, obj):
        try:
            self.driver.device_driver.update_member(obj, old_obj)
            self.successful_completion(context, obj)
        except Exception:
            self.failed_completion(context, obj)

    def delete(self, context, obj):
        try:
            self.driver.device_driver.delete_member(obj)
        except Exception:
            # Ignore the exception
            pass

        self.successful_completion(context, obj, delete=True)


class BrocadeHealthMonitorManager(driver_base.BaseHealthMonitorManager):
    def create(self, context, obj):
        try:
            self.driver.device_driver.create_healthmonitor(obj)
            self.successful_completion(context, obj)
        except Exception:
            self.failed_completion(context, obj)

    def update(self, context, old_obj, obj):
        try:
            self.driver.device_driver.update_healthmonitor(obj, old_obj)
            self.successful_completion(context, obj)
        except Exception:
            self.failed_completion(context, obj)

    def delete(self, context, obj):
        try:
            self.driver.device_driver.delete_healthmonitor(obj)
        except Exception:
            # Ignore the exception
            pass

        self.successful_completion(context, obj, delete=True)



#  Copyright 2015, Shane McGough, KEMPtechnologies
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import sys

import mock
from neutron_lib import context

from neutron_lbaas.tests.unit.db.loadbalancer import test_db_loadbalancerv2

with mock.patch.dict(sys.modules, {'kemptech_openstack_lbaas': mock.Mock()}):
    from neutron_lbaas.drivers.kemptechnologies import driver_v2


class FakeModel(object):
    def __init__(self, id):
        self.id = id
        self.address = '1.1.1.1'
        self.tenant_id = "copying-pre-existing-work-is-easy"


class ManagerTest(object):
    def __init__(self, parent, manager, model, mocked_root):
        self.parent = parent
        self.context = parent.context
        self.driver = parent.driver
        self.manager = manager
        self.model = model
        self.mocked_root = mocked_root

        self.create(model)
        self.update(model, model)
        self.delete(model)

    def create(self, model):
        self.manager.create(self.context, model)
        self.mocked_root.create.assert_called_with(self.context, model)

    def update(self, old_model, model):
        self.manager.update(self.context, old_model, model)
        self.mocked_root.update.assert_called_with(self.context,
                                                   old_model, model)

    def delete(self, model):
        self.manager.delete(self.context, model)
        self.mocked_root.delete.assert_called_with(self.context, model)

    def refresh(self):
        self.manager.refresh(self.context, self.model)
        self.mocked_root.refresh.assert_called_with(self.context, self.model)

    def stats(self):
        self.manager.stats(self.context, self.model)
        self.mocked_root.stats.assert_called_with(self.context, self.model)


class TestKempLoadMasterDriver(test_db_loadbalancerv2.LbaasPluginDbTestCase):

    def setUp(self):
        super(TestKempLoadMasterDriver, self).setUp()
        self.context = context.get_admin_context()
        self.plugin = mock.Mock()
        self.driver = driver_v2.KempLoadMasterDriver(self.plugin)
        self.driver.kemptech = mock.Mock()

    def test_load_balancer_ops(self):
        m = ManagerTest(self, self.driver.load_balancer,
                        FakeModel("load_balancer-kemptech"),
                        self.driver.kemptech.load_balancer)
        m.refresh()
        m.stats()

    def test_listener_ops(self):
        ManagerTest(self, self.driver.listener, FakeModel("listener-kemptech"),
                    self.driver.kemptech.listener)

    def test_pool_ops(self):
        ManagerTest(self, self.driver.pool, FakeModel("pool-kemptech"),
                    self.driver.kemptech.pool)

    def test_member_ops(self):
        ManagerTest(self, self.driver.member, FakeModel("member-kemptech"),
                    self.driver.kemptech.member)

    def test_health_monitor_ops(self):
        ManagerTest(self, self.driver.health_monitor,
                    FakeModel("health_monitor-kemptech"),
                    self.driver.kemptech.health_monitor)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2016 Rackspace Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import utils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as ex

from neutron_lbaas.tests.tempest.v2.api import base


class ListenersTestJSON(base.BaseAdminTestCase):

    """
    Tests the listener creation operation in admin scope in the
    Neutron-LBaaS API using the REST client for Listeners:

    """

    @classmethod
    def resource_setup(cls):
        super(ListenersTestJSON, cls).resource_setup()
        if not utils.is_extension_enabled('lbaasv2', 'network'):
            msg = "lbaas extension not enabled."
            raise cls.skipException(msg)
        network_name = data_utils.rand_name('network-')
        cls.network = cls.create_network(network_name)
        cls.subnet = cls.create_subnet(cls.network)
        cls.create_lb_kwargs = {'tenant_id': cls.subnet['tenant_id'],
                                'vip_subnet_id': cls.subnet['id']}
        cls.load_balancer = cls._create_active_load_balancer(
            **cls.create_lb_kwargs)
        cls.port = 80
        cls.load_balancer_id = cls.load_balancer['id']
        cls.create_listener_kwargs = {'loadbalancer_id': cls.load_balancer_id,
                                      'protocol': cls.listener_protocol,
                                      'protocol_port': cls.port}
        cls.listener = cls._create_listener(
            **cls.create_listener_kwargs)
        cls.listener_id = cls.listener['id']

    @classmethod
    def resource_cleanup(cls):
        super(ListenersTestJSON, cls).resource_cleanup()

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_create_listener_empty_tenant_id(self):
        """Test create listener with an empty tenant id should fail"""
        create_new_listener_kwargs = self.create_listener_kwargs
        create_new_listener_kwargs['protocol_port'] = 8081
        create_new_listener_kwargs['tenant_id'] = ""
        self.assertRaises(ex.BadRequest,
                          self._create_listener,
                          **create_new_listener_kwargs)
        self._check_status_tree(
            load_balancer_id=self.load_balancer_id,
            listener_ids=[self.listener_id])

    @decorators.skip_because(bug="1468457")
    def test_create_listener_invalid_tenant_id(self):
        """Test create listener with an invalid tenant id"""
        create_new_listener_kwargs = self.create_listener_kwargs
        create_new_listener_kwargs['protocol_port'] = 8081
        create_new_listener_kwargs['tenant_id'] = "&^%123"
        new_listener = self._create_listener(
            **create_new_listener_kwargs)
        new_listener_id = new_listener['id']
        self.addCleanup(self._delete_listener, new_listener_id)
        self._check_status_tree(
            load_balancer_id=self.load_balancer_id,
            listener_ids=[self.listener_id, new_listener_id])
        listener = self.listeners_client.get_listener(
            new_listener_id)
        self.assertEqual(new_listener, listener)

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='smoke')
    def test_create_listener_missing_tenant_id(self):
        """Test create listener with an missing tenant id.

        Verify that creating a listener in admin scope with
        a missing tenant_id creates the listener with admin
        tenant_id.
        """
        create_new_listener_kwargs = self.create_listener_kwargs
        create_new_listener_kwargs['protocol_port'] = 8081
        admin_listener = self._create_listener(
            **create_new_listener_kwargs)
        admin_listener_id = admin_listener['id']
        self.addCleanup(self._delete_listener, admin_listener_id)
        self._check_status_tree(
            load_balancer_id=self.load_balancer_id,
            listener_ids=[self.listener_id, admin_listener_id])
        listener = self.listeners_client.get_listener(
            admin_listener_id)
        self.assertEqual(admin_listener, listener)
        self.assertEqual(admin_listener.get('tenant_id'),
                         listener.get('tenant_id'))



# Copyright 2016-2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import f5lbaasdriver
from f5lbaasdriver.v2.bigip.driver_v2 import F5DriverV2

from oslo_log import log as logging

from neutron_lbaas.drivers import driver_base

VERSION = "1.0.0"
LOG = logging.getLogger(__name__)


class UndefinedEnvironment(Exception):
    pass


class F5LBaaSV2Driver(driver_base.LoadBalancerBaseDriver):

    def __init__(self, plugin, env='Project'):
        super(F5LBaaSV2Driver, self).__init__(plugin)

        self.load_balancer = LoadBalancerManager(self)
        self.listener = ListenerManager(self)
        self.pool = PoolManager(self)
        self.member = MemberManager(self)
        self.health_monitor = HealthMonitorManager(self)
        self.l7policy = L7PolicyManager(self)
        self.l7rule = L7RuleManager(self)

        if not env:
            msg = "F5LBaaSV2Driver cannot be intialized because the "\
                "environment is not defined. To set the environment, edit "\
                "neutron_lbaas.conf and append the environment name to the "\
                "service_provider class name."
            LOG.debug(msg)
            raise UndefinedEnvironment(msg)

        LOG.debug("F5LBaaSV2Driver: initializing, version=%s, impl=%s, env=%s"
                  % (VERSION, f5lbaasdriver.__version__, env))

        self.f5 = F5DriverV2(plugin, env)


class F5LBaaSV2DriverTest(F5LBaaSV2Driver):

    def __init__(self, plugin, env='Test'):
        super(F5LBaaSV2DriverTest, self).__init__(plugin, env)

        LOG.debug(
            "F5LBaaSV2DriverTest: initializing, version=%s, f5=%s, env=%s"
            % (VERSION, f5lbaasdriver.__version__, env))


class F5LBaaSV2DriverProject(F5LBaaSV2Driver):

    def __init__(self, plugin, env='Project'):
        super(F5LBaaSV2DriverProject, self).__init__(plugin, env)

        LOG.debug(
            "F5LBaaSV2DriverProject: initializing, version=%s, f5=%s, env=%s"
            % (VERSION, f5lbaasdriver.__version__, env))


class LoadBalancerManager(driver_base.BaseLoadBalancerManager):

    def create(self, context, lb):
        self.driver.f5.loadbalancer.create(context, lb)

    def update(self, context, old_lb, lb):
        self.driver.f5.loadbalancer.update(context, old_lb, lb)

    def delete(self, context, lb):
        self.driver.f5.loadbalancer.delete(context, lb)

    def refresh(self, context, lb):
        self.driver.f5.loadbalancer.refresh(context, lb)

    def stats(self, context, lb):
        return self.driver.f5.loadbalancer.stats(context, lb)


class ListenerManager(driver_base.BaseListenerManager):

    def create(self, context, listener):
        self.driver.f5.listener.create(context, listener)

    def update(self, context, old_listener, listener):
        self.driver.f5.listener.update(context, old_listener, listener)

    def delete(self, context, listener):
        self.driver.f5.listener.delete(context, listener)


class PoolManager(driver_base.BasePoolManager):

    def create(self, context, pool):
        self.driver.f5.pool.create(context, pool)

    def update(self, context, old_pool, pool):
        self.driver.f5.pool.update(context, old_pool, pool)

    def delete(self, context, pool):
        self.driver.f5.pool.delete(context, pool)


class MemberManager(driver_base.BaseMemberManager):

    def create(self, context, member):
        self.driver.f5.member.create(context, member)

    def update(self, context, old_member, member):
        self.driver.f5.member.update(context, old_member, member)

    def delete(self, context, member):
        self.driver.f5.member.delete(context, member)


class HealthMonitorManager(driver_base.BaseHealthMonitorManager):

    def create(self, context, health_monitor):
        self.driver.f5.healthmonitor.create(context, health_monitor)

    def update(self, context, old_health_monitor, health_monitor):
        self.driver.f5.healthmonitor.update(context, old_health_monitor,
                                   health_monitor)

    def delete(self, context, health_monitor):
        self.driver.f5.healthmonitor.delete(context, health_monitor)


class L7PolicyManager(driver_base.BaseL7PolicyManager):

    def create(self, context, l7policy):
        self.driver.f5.l7policy.create(context, l7policy)

    def update(self, context, old_l7policy, l7policy):
        self.driver.f5.l7policy.update(context, old_l7policy, l7policy)

    def delete(self, context, l7policy):
        self.driver.f5.l7policy.delete(context, l7policy)


class L7RuleManager(driver_base.BaseL7RuleManager):

    def create(self, context, l7rule):
        self.driver.f5.l7rule.create(context, l7rule)

    def update(self, context, old_l7rule, l7rule):
        self.driver.f5.l7rule.update(context, old_l7rule, l7rule)

    def delete(self, context, l7rule):
        self.driver.f5.l7rule.delete(context, l7rule)



# Copyright 2016 Rackspace Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import time

from six.moves.urllib import parse
from tempest.lib.common import rest_client
from tempest.lib.common.utils import test_utils
from tempest.lib import exceptions as lib_exc


class NetworkClientJSON(rest_client.RestClient):

    """
    Tempest REST client for Neutron. Uses v2 of the Neutron API, since the
    V1 API has been removed from the code base.

    Implements create, delete, update, list and show for the basic Neutron
    abstractions (networks, sub-networks, routers, ports and floating IP):

    Implements add/remove interface to router using subnet ID / port ID

    It also implements list, show, update and reset for OpenStack Networking
    quotas
    """

    version = '2.0'
    uri_prefix = "v2.0"

    def get_uri(self, plural_name):
        # get service prefix from resource name

        # The following list represents resource names that do not require
        # changing underscore to a hyphen
        hyphen_exceptions = ["health_monitors", "firewall_rules",
                             "firewall_policies"]
        # the following map is used to construct proper URI
        # for the given neutron resource
        service_resource_prefix_map = {
            'networks': '',
            'subnets': '',
            'subnetpools': '',
            'ports': '',
            'pools': 'lb',
            'vips': 'lb',
            'health_monitors': 'lb',
            'members': 'lb',
            'ipsecpolicies': 'vpn',
            'vpnservices': 'vpn',
            'ikepolicies': 'vpn',
            'ipsec-site-connections': 'vpn',
            'metering_labels': 'metering',
            'metering_label_rules': 'metering',
            'firewall_rules': 'fw',
            'firewall_policies': 'fw',
            'firewalls': 'fw'
        }
        service_prefix = service_resource_prefix_map.get(
            plural_name)
        if plural_name not in hyphen_exceptions:
            plural_name = plural_name.replace("_", "-")
        if service_prefix:
            uri = '%s/%s/%s' % (self.uri_prefix, service_prefix,
                                plural_name)
        else:
            uri = '%s/%s' % (self.uri_prefix, plural_name)
        return uri

    def pluralize(self, resource_name):
        # get plural from map or just add 's'

        # map from resource name to a plural name
        # needed only for those which can't be constructed as name + 's'
        resource_plural_map = {
            'security_groups': 'security_groups',
            'security_group_rules': 'security_group_rules',
            'ipsecpolicy': 'ipsecpolicies',
            'ikepolicy': 'ikepolicies',
            'ipsec_site_connection': 'ipsec-site-connections',
            'quotas': 'quotas',
            'firewall_policy': 'firewall_policies'
        }
        return resource_plural_map.get(resource_name, resource_name + 's')

    def _lister(self, plural_name):
        def _list(**filters):
            uri = self.get_uri(plural_name)
            if filters:
                uri += '?' + parse.urlencode(filters, doseq=1)
            resp, body = self.get(uri)
            result = {plural_name: self.deserialize_list(body)}
            self.expected_success(200, resp.status)
            return rest_client.ResponseBody(resp, result)

        return _list

    def _deleter(self, resource_name):
        def _delete(resource_id):
            plural = self.pluralize(resource_name)
            uri = '%s/%s' % (self.get_uri(plural), resource_id)
            resp, body = self.delete(uri)
            self.expected_success(204, resp.status)
            return rest_client.ResponseBody(resp, body)

        return _delete

    def _shower(self, resource_name):
        def _show(resource_id, **fields):
            # fields is a dict which key is 'fields' and value is a
            # list of field's name. An example:
            # {'fields': ['id', 'name']}
            plural = self.pluralize(resource_name)
            uri = '%s/%s' % (self.get_uri(plural), resource_id)
            if fields:
                uri += '?' + parse.urlencode(fields, doseq=1)
            resp, body = self.get(uri)
            body = self.deserialize_single(body)
            self.expected_success(200, resp.status)
            return rest_client.ResponseBody(resp, body)

        return _show

    def _creater(self, resource_name):
        def _create(**kwargs):
            plural = self.pluralize(resource_name)
            uri = self.get_uri(plural)
            post_data = self.serialize({resource_name: kwargs})
            resp, body = self.post(uri, post_data)
            body = self.deserialize_single(body)
            self.expected_success(201, resp.status)
            return rest_client.ResponseBody(resp, body)

        return _create

    def _updater(self, resource_name):
        def _update(res_id, **kwargs):
            plural = self.pluralize(resource_name)
            uri = '%s/%s' % (self.get_uri(plural), res_id)
            post_data = self.serialize({resource_name: kwargs})
            resp, body = self.put(uri, post_data)
            body = self.deserialize_single(body)
            self.expected_success(200, resp.status)
            return rest_client.ResponseBody(resp, body)

        return _update

    def __getattr__(self, name):
        method_prefixes = ["list_", "delete_", "show_", "create_", "update_"]
        method_functors = [self._lister,
                           self._deleter,
                           self._shower,
                           self._creater,
                           self._updater]
        for index, prefix in enumerate(method_prefixes):
            prefix_len = len(prefix)
            if name[:prefix_len] == prefix:
                return method_functors[index](name[prefix_len:])
        raise AttributeError(name)

    # Subnetpool methods
    def create_subnetpool(self, post_data):
        body = self.serialize_list(post_data, "subnetpools", "subnetpool")
        uri = self.get_uri("subnetpools")
        resp, body = self.post(uri, body)
        body = {'subnetpool': self.deserialize_list(body)}
        self.expected_success(201, resp.status)
        return rest_client.ResponseBody(resp, body)

    def get_subnetpool(self, id):
        uri = self.get_uri("subnetpools")
        subnetpool_uri = '%s/%s' % (uri, id)
        resp, body = self.get(subnetpool_uri)
        body = {'subnetpool': self.deserialize_list(body)}
        self.expected_success(200, resp.status)
        return rest_client.ResponseBody(resp, body)

    def delete_subnetpool(self, id):
        uri = self.get_uri("subnetpools")
        subnetpool_uri = '%s/%s' % (uri, id)
        resp, body = self.delete(subnetpool_uri)
        self.expected_success(204, resp.status)
        return rest_client.ResponseBody(resp, body)

    def list_subnetpools(self):
        uri = self.get_uri("subnetpools")
        resp, body = self.get(uri)
        body = {'subnetpools': self.deserialize_list(body)}
        self.expected_success(200, resp.status)
        return rest_client.ResponseBody(resp, body)

    def update_subnetpool(self, id, post_data):
        body = self.serialize_list(post_data, "subnetpools", "subnetpool")
        uri = self.get_uri("subnetpools")
        subnetpool_uri = '%s/%s' % (uri, id)
        resp, body = self.put(subnetpool_uri, body)
        body = {'subnetpool': self.deserialize_list(body)}
        self.expected_success(200, resp.status)
        return rest_client.ResponseBody(resp, body)

    # Common methods that are hard to automate
    def create_bulk_network(self, names, shared=False):
        network_list = [{'name': name, 'shared': shared} for name in names]
        post_data = {'networks': network_list}
        body = self.serialize_list(post_data, "networks", "network")
        uri = self.get_uri("networks")
        resp, body = self.post(uri, body)
        body = {'networks': self.deserialize_list(body)}
        self.expected_success(201, resp.status)
        return rest_client.ResponseBody(resp, body)

    def create_bulk_subnet(self, subnet_list):
        post_data = {'subnets': subnet_list}
        body = self.serialize_list(post_data, 'subnets', 'subnet')
        uri = self.get_uri('subnets')
        resp, body = self.post(uri, body)
        body = {'subnets': self.deserialize_list(body)}
        self.expected_success(201, resp.status)
        return rest_client.ResponseBody(resp, body)

    def create_bulk_port(self, port_list):
        post_data = {'ports': port_list}
        body = self.serialize_list(post_data, 'ports', 'port')
        uri = self.get_uri('ports')
        resp, body = self.post(uri, body)
        body = {'ports': self.deserialize_list(body)}
        self.expected_success(201, resp.status)
        return rest_client.ResponseBody(resp, body)

    def wait_for_resource_deletion(self, resource_type, id):
        """Waits for a resource to be deleted."""
        start_time = int(time.time())
        while True:
            if self.is_resource_deleted(resource_type, id):
                return
            if int(time.time()) - start_time >= self.build_timeout:
                raise lib_exc.TimeoutException
            time.sleep(self.build_interval)

    def is_resource_deleted(self, resource_type, id):
        method = 'show_' + resource_type
        try:
            getattr(self, method)(id)
        except AttributeError:
            raise Exception("Unknown resource type %s " % resource_type)
        except lib_exc.NotFound:
            return True
        return False

    def wait_for_resource_status(self, fetch, status, interval=None,
                                 timeout=None):
        """
        @summary: Waits for a network resource to reach a status
        @param fetch: the callable to be used to query the resource status
        @type fecth: callable that takes no parameters and returns the resource
        @param status: the status that the resource has to reach
        @type status: String
        @param interval: the number of seconds to wait between each status
          query
        @type interval: Integer
        @param timeout: the maximum number of seconds to wait for the resource
          to reach the desired status
        @type timeout: Integer
        """
        if not interval:
            interval = self.build_interval
        if not timeout:
            timeout = self.build_timeout
        start_time = time.time()

        while time.time() - start_time <= timeout:
            resource = fetch()
            if resource['status'] == status:
                return
            time.sleep(interval)

        # At this point, the wait has timed out
        message = 'Resource %s' % (str(resource))
        message += ' failed to reach status %s' % status
        message += ' (current: %s)' % resource['status']
        message += ' within the required time %s' % timeout
        caller = test_utils.find_test_caller()
        if caller:
            message = '(%s) %s' % (caller, message)
        raise lib_exc.TimeoutException(message)

    def deserialize_single(self, body):
        return json.loads(body)

    def deserialize_list(self, body):
        res = json.loads(body)
        # expecting response in form
        # {'resources': [ res1, res2] } => when pagination disabled
        # {'resources': [..], 'resources_links': {}} => if pagination enabled
        for k in res.keys():
            if k.endswith("_links"):
                continue
            return res[k]

    def serialize(self, data):
        return json.dumps(data)

    def serialize_list(self, data, root=None, item=None):
        return self.serialize(data)

    def update_quotas(self, tenant_id, **kwargs):
        put_body = {'quota': kwargs}
        body = json.dumps(put_body)
        uri = '%s/quotas/%s' % (self.uri_prefix, tenant_id)
        resp, body = self.put(uri, body)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body['quota'])

    def reset_quotas(self, tenant_id):
        uri = '%s/quotas/%s' % (self.uri_prefix, tenant_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return rest_client.ResponseBody(resp, body)

    def create_router(self, name, admin_state_up=True, **kwargs):
        post_body = {'router': kwargs}
        post_body['router']['name'] = name
        post_body['router']['admin_state_up'] = admin_state_up
        body = json.dumps(post_body)
        uri = '%s/routers' % (self.uri_prefix)
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def _update_router(self, router_id, set_enable_snat, **kwargs):
        uri = '%s/routers/%s' % (self.uri_prefix, router_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        update_body = {}
        update_body['name'] = kwargs.get('name', body['router']['name'])
        update_body['admin_state_up'] = kwargs.get(
            'admin_state_up', body['router']['admin_state_up'])
        cur_gw_info = body['router']['external_gateway_info']
        if cur_gw_info:
            # TODO(kevinbenton): setting the external gateway info is not
            # allowed for a regular tenant. If the ability to update is also
            # merged, a test case for this will need to be added similar to
            # the SNAT case.
            cur_gw_info.pop('external_fixed_ips', None)
            if not set_enable_snat:
                cur_gw_info.pop('enable_snat', None)
        update_body['external_gateway_info'] = kwargs.get(
            'external_gateway_info', body['router']['external_gateway_info'])
        if 'distributed' in kwargs:
            update_body['distributed'] = kwargs['distributed']
        update_body = dict(router=update_body)
        update_body = json.dumps(update_body)
        resp, body = self.put(uri, update_body)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def update_router(self, router_id, **kwargs):
        """Update a router leaving enable_snat to its default value."""
        # If external_gateway_info contains enable_snat the request will fail
        # with 404 unless executed with admin client, and therefore we instruct
        # _update_router to not set this attribute
        # NOTE(salv-orlando): The above applies as long as Neutron's default
        # policy is to restrict enable_snat usage to admins only.
        return self._update_router(router_id, set_enable_snat=False, **kwargs)

    def update_router_with_snat_gw_info(self, router_id, **kwargs):
        """Update a router passing also the enable_snat attribute.

        This method must be execute with admin credentials, otherwise the API
        call will return a 404 error.
        """
        return self._update_router(router_id, set_enable_snat=True, **kwargs)

    def add_router_interface_with_subnet_id(self, router_id, subnet_id):
        uri = '%s/routers/%s/add_router_interface' % (self.uri_prefix,
                                                      router_id)
        update_body = {"subnet_id": subnet_id}
        update_body = json.dumps(update_body)
        resp, body = self.put(uri, update_body)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def add_router_interface_with_port_id(self, router_id, port_id):
        uri = '%s/routers/%s/add_router_interface' % (self.uri_prefix,
                                                      router_id)
        update_body = {"port_id": port_id}
        update_body = json.dumps(update_body)
        resp, body = self.put(uri, update_body)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def remove_router_interface_with_subnet_id(self, router_id, subnet_id):
        uri = '%s/routers/%s/remove_router_interface' % (self.uri_prefix,
                                                         router_id)
        update_body = {"subnet_id": subnet_id}
        update_body = json.dumps(update_body)
        resp, body = self.put(uri, update_body)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def remove_router_interface_with_port_id(self, router_id, port_id):
        uri = '%s/routers/%s/remove_router_interface' % (self.uri_prefix,
                                                         router_id)
        update_body = {"port_id": port_id}
        update_body = json.dumps(update_body)
        resp, body = self.put(uri, update_body)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def associate_health_monitor_with_pool(self, health_monitor_id,
                                           pool_id):
        post_body = {
            "health_monitor": {
                "id": health_monitor_id,
            }
        }
        body = json.dumps(post_body)
        uri = '%s/lb/pools/%s/health_monitors' % (self.uri_prefix,
                                                  pool_id)
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def disassociate_health_monitor_with_pool(self, health_monitor_id,
                                              pool_id):
        uri = '%s/lb/pools/%s/health_monitors/%s' % (self.uri_prefix, pool_id,
                                                     health_monitor_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return rest_client.ResponseBody(resp, body)

    def list_router_interfaces(self, uuid):
        uri = '%s/ports?device_id=%s' % (self.uri_prefix, uuid)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def update_agent(self, agent_id, agent_info):
        """
        :param agent_info: Agent update information.
        E.g {"admin_state_up": True}
        """
        uri = '%s/agents/%s' % (self.uri_prefix, agent_id)
        agent = {"agent": agent_info}
        body = json.dumps(agent)
        resp, body = self.put(uri, body)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def list_pools_hosted_by_one_lbaas_agent(self, agent_id):
        uri = '%s/agents/%s/loadbalancer-pools' % (self.uri_prefix, agent_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def show_lbaas_agent_hosting_pool(self, pool_id):
        uri = ('%s/lb/pools/%s/loadbalancer-agent' %
               (self.uri_prefix, pool_id))
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def list_routers_on_l3_agent(self, agent_id):
        uri = '%s/agents/%s/l3-routers' % (self.uri_prefix, agent_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def list_l3_agents_hosting_router(self, router_id):
        uri = '%s/routers/%s/l3-agents' % (self.uri_prefix, router_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def add_router_to_l3_agent(self, agent_id, router_id):
        uri = '%s/agents/%s/l3-routers' % (self.uri_prefix, agent_id)
        post_body = {"router_id": router_id}
        body = json.dumps(post_body)
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def remove_router_from_l3_agent(self, agent_id, router_id):
        uri = '%s/agents/%s/l3-routers/%s' % (
            self.uri_prefix, agent_id, router_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return rest_client.ResponseBody(resp, body)

    def list_dhcp_agent_hosting_network(self, network_id):
        uri = '%s/networks/%s/dhcp-agents' % (self.uri_prefix, network_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def list_networks_hosted_by_one_dhcp_agent(self, agent_id):
        uri = '%s/agents/%s/dhcp-networks' % (self.uri_prefix, agent_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def remove_network_from_dhcp_agent(self, agent_id, network_id):
        uri = '%s/agents/%s/dhcp-networks/%s' % (self.uri_prefix, agent_id,
                                                 network_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return rest_client.ResponseBody(resp, body)

    def create_ikepolicy(self, name, **kwargs):
        post_body = {
            "ikepolicy": {
                "name": name,
            }
        }
        for key, val in kwargs.items():
            post_body['ikepolicy'][key] = val
        body = json.dumps(post_body)
        uri = '%s/vpn/ikepolicies' % (self.uri_prefix)
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def update_extra_routes(self, router_id, nexthop, destination):
        uri = '%s/routers/%s' % (self.uri_prefix, router_id)
        put_body = {
            'router': {
                'routes': [{'nexthop': nexthop,
                            "destination": destination}]
            }
        }
        body = json.dumps(put_body)
        resp, body = self.put(uri, body)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def delete_extra_routes(self, router_id):
        uri = '%s/routers/%s' % (self.uri_prefix, router_id)
        null_routes = None
        put_body = {
            'router': {
                'routes': null_routes
            }
        }
        body = json.dumps(put_body)
        resp, body = self.put(uri, body)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def list_lb_pool_stats(self, pool_id):
        uri = '%s/lb/pools/%s/stats' % (self.uri_prefix, pool_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def add_dhcp_agent_to_network(self, agent_id, network_id):
        post_body = {'network_id': network_id}
        body = json.dumps(post_body)
        uri = '%s/agents/%s/dhcp-networks' % (self.uri_prefix, agent_id)
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def insert_firewall_rule_in_policy(self, firewall_policy_id,
                                       firewall_rule_id, insert_after="",
                                       insert_before=""):
        uri = '%s/fw/firewall_policies/%s/insert_rule' % (self.uri_prefix,
                                                          firewall_policy_id)
        body = {
            "firewall_rule_id": firewall_rule_id,
            "insert_after": insert_after,
            "insert_before": insert_before
        }
        body = json.dumps(body)
        resp, body = self.put(uri, body)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)

    def remove_firewall_rule_from_policy(self, firewall_policy_id,
                                         firewall_rule_id):
        uri = '%s/fw/firewall_policies/%s/remove_rule' % (self.uri_prefix,
                                                          firewall_policy_id)
        update_body = {"firewall_rule_id": firewall_rule_id}
        update_body = json.dumps(update_body)
        resp, body = self.put(uri, update_body)
        self.expected_success(200, resp.status)
        body = json.loads(body)
        return rest_client.ResponseBody(resp, body)



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add_l7_tables

Revision ID: 3543deab1547
Revises: 6aee0434f911
Create Date: 2015-02-05 10:50:15.606420

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3543deab1547'
down_revision = '6aee0434f911'

l7rule_type = sa.Enum("HOST_NAME", "PATH", "FILE_TYPE", "HEADER", "COOKIE",
                      name="l7rule_typesv2")
l7rule_compare_type = sa.Enum("REGEX", "STARTS_WITH", "ENDS_WITH", "CONTAINS",
                              "EQUAL_TO", name="l7rule_compare_typesv2")
l7policy_action_type = sa.Enum("REJECT", "REDIRECT_TO_URL", "REDIRECT_TO_POOL",
                               name="l7policy_action_typesv2")


def upgrade():
    op.create_table(
        u'lbaas_l7policies',
        sa.Column(u'tenant_id', sa.String(255), nullable=True),
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'name', sa.String(255), nullable=True),
        sa.Column(u'description', sa.String(255), nullable=True),
        sa.Column(u'listener_id', sa.String(36), nullable=False),
        sa.Column(u'action', l7policy_action_type, nullable=False),
        sa.Column(u'redirect_pool_id', sa.String(36), nullable=True),
        sa.Column(u'redirect_url', sa.String(255), nullable=True),
        sa.Column(u'position', sa.Integer, nullable=False),
        sa.Column(u'provisioning_status', sa.String(16), nullable=False),
        sa.Column(u'admin_state_up', sa.Boolean(), nullable=False),

        sa.PrimaryKeyConstraint(u'id'),
        sa.ForeignKeyConstraint([u'listener_id'],
                                [u'lbaas_listeners.id']),
        sa.ForeignKeyConstraint([u'redirect_pool_id'],
                                [u'lbaas_pools.id'])
    )

    op.create_table(
        u'lbaas_l7rules',
        sa.Column(u'tenant_id', sa.String(255), nullable=True),
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'l7policy_id', sa.String(36), nullable=False),
        sa.Column(u'type', l7rule_type, nullable=False),
        sa.Column(u'compare_type', l7rule_compare_type, nullable=False),
        sa.Column(u'invert', sa.Boolean(), nullable=False),
        sa.Column(u'key', sa.String(255), nullable=True),
        sa.Column(u'value', sa.String(255), nullable=False),
        sa.Column(u'provisioning_status', sa.String(16), nullable=False),
        sa.Column(u'admin_state_up', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint(u'id'),
        sa.ForeignKeyConstraint([u'l7policy_id'],
                                [u'lbaas_l7policies.id'])
    )



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""lbaasv2 TLS

Revision ID: lbaasv2_tls
Revises: 364f9b6064f0
Create Date: 2015-01-18 10:00:00

"""

from alembic import op
import sqlalchemy as sa

from neutron.db import migration

# revision identifiers, used by Alembic.
revision = 'lbaasv2_tls'
down_revision = '364f9b6064f0'


old_listener_protocols = sa.Enum("HTTP", "HTTPS", "TCP",
                             name="listener_protocolsv2")
new_listener_protocols = sa.Enum("HTTP", "HTTPS", "TCP", "TERMINATED_HTTPS",
                             name="listener_protocolsv2")


def upgrade():
    migration.alter_enum('lbaas_listeners', 'protocol', new_listener_protocols,
                         nullable=False)
    op.create_table(
        u'lbaas_sni',
        sa.Column(u'listener_id', sa.String(36), nullable=False),
        sa.Column(u'tls_container_id', sa.String(128), nullable=False),
        sa.Column(u'position', sa.Integer),
        sa.ForeignKeyConstraint(['listener_id'], [u'lbaas_listeners.id'], ),
        sa.PrimaryKeyConstraint(u'listener_id', u'tls_container_id')
    )

    op.add_column('lbaas_listeners',
                  sa.Column(u'default_tls_container_id', sa.String(128),
                            nullable=True))



# Copyright 2015, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from datetime import datetime
from functools import wraps
import threading
import time

from neutron_lib import context as ncontext
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_service import service
from oslo_utils import excutils
import requests

from neutron_lbaas._i18n import _
from neutron_lbaas.common import exceptions
from neutron_lbaas.common import keystone
from neutron_lbaas.drivers import driver_base
from neutron_lbaas.drivers.octavia import octavia_messaging_consumer
from neutron_lbaas.services.loadbalancer import constants

LOG = logging.getLogger(__name__)
VERSION = "1.0.1"

OPTS = [
    cfg.StrOpt(
        'base_url',
        default='http://127.0.0.1:9876',
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        help=_('URL of Octavia controller root'),
    ),
    cfg.IntOpt(
        'request_poll_interval',
        default=3,
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        help=_('Interval in seconds to poll octavia when an entity is created,'
               ' updated, or deleted.')
    ),
    cfg.IntOpt(
        'request_poll_timeout',
        default=100,
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        help=_('Time to stop polling octavia when a status of an entity does '
               'not change.')
    ),
    cfg.BoolOpt(
        'allocates_vip',
        default=False,
        deprecated_for_removal=True,
        deprecated_since='Queens',
        deprecated_reason='The neutron-lbaas project is now deprecated. '
                          'See: https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                          'Deprecation',
        help=_('True if Octavia will be responsible for allocating the VIP.'
               ' False if neutron-lbaas will allocate it and pass to Octavia.')
    )
]

cfg.CONF.register_opts(OPTS, 'octavia')


def thread_op(manager, entity, delete=False, lb_create=False):
    context = ncontext.get_admin_context()
    poll_interval = cfg.CONF.octavia.request_poll_interval
    poll_timeout = cfg.CONF.octavia.request_poll_timeout
    start_dt = datetime.now()
    prov_status = None
    while (datetime.now() - start_dt).seconds < poll_timeout:
        octavia_lb = manager.driver.load_balancer.get(entity.root_loadbalancer)
        prov_status = octavia_lb.get('provisioning_status')
        LOG.debug("Octavia reports load balancer {0} has provisioning status "
                  "of {1}".format(entity.root_loadbalancer.id, prov_status))
        if prov_status == 'ACTIVE' or prov_status == 'DELETED':
            kwargs = {'delete': delete}
            if manager.driver.allocates_vip and lb_create:
                kwargs['lb_create'] = lb_create
                # TODO(blogan): drop fk constraint on vip_port_id to ports
                # table because the port can't be removed unless the load
                # balancer has been deleted.  Until then we won't populate the
                # vip_port_id field.
                # entity.vip_port_id = octavia_lb.get('vip').get('port_id')
                entity.vip_address = octavia_lb.get('vip').get('ip_address')
            manager.successful_completion(context, entity, **kwargs)
            return
        elif prov_status == 'ERROR':
            manager.failed_completion(context, entity)
            return
        time.sleep(poll_interval)
    LOG.warning("Timeout has expired for load balancer {0} to complete an "
              "operation.  The last reported status was "
              "{1}".format(entity.root_loadbalancer.id, prov_status))
    manager.failed_completion(context, entity)


# A decorator for wrapping driver operations, which will automatically
# set the neutron object's status based on whether it sees an exception

def async_op(func):
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        d = (func.__name__ == 'delete' or func.__name__ == 'delete_cascade')
        lb_create = ((func.__name__ == 'create') and
                     isinstance(args[0], LoadBalancerManager))
        try:
            r = func(*args, **kwargs)
            thread = threading.Thread(target=thread_op,
                                      args=(args[0], args[2]),
                                      kwargs={'delete': d,
                                              'lb_create': lb_create})
            thread.setDaemon(True)
            thread.start()
            return r
        except Exception:
            with excutils.save_and_reraise_exception():
                args[0].failed_completion(args[1], args[2])
    return func_wrapper


class OctaviaRequest(object):

    def __init__(self, base_url, auth_session):
        self.base_url = base_url
        self.auth_session = auth_session

    def request(self, method, url, args=None, headers=None):
        if args:
            args = jsonutils.dumps(args)

        if not headers or not headers.get('X-Auth-Token'):
            headers = headers or {
                'Content-type': 'application/json',
            }
            headers['X-Auth-Token'] = self.auth_session.get_token()

        LOG.debug("url = %s", '%s%s' % (self.base_url, str(url)))
        LOG.debug("args = %s", args)
        r = requests.request(method,
                             '%s%s' % (self.base_url, str(url)),
                             data=args,
                             headers=headers)
        LOG.debug("Octavia Response Code: {0}".format(r.status_code))
        LOG.debug("Octavia Response Body: {0}".format(r.content))
        LOG.debug("Octavia Response Headers: {0}".format(r.headers))
        # We need to raise Octavia errors up to neutron API.
        try:
            fault_string = jsonutils.loads(r.content)['faultstring']
        except Exception:
            fault_string = "Unknown Octavia error."
        if r.status_code == 400:
            raise exceptions.BadRequestException(fault_string=fault_string)
        elif r.status_code == 401:
            raise exceptions.NotAuthorizedException(fault_string=fault_string)
        elif r.status_code == 403:
            raise exceptions.NotAuthorizedException(fault_string=fault_string)
        elif r.status_code == 404:
            raise exceptions.NotFoundException(fault_string=fault_string)
        elif r.status_code == 409:
            raise exceptions.ConflictException(fault_string=fault_string)
        elif r.status_code == 500:
            raise exceptions.UnknownException(fault_string=fault_string)
        elif r.status_code == 503:
            raise exceptions.ServiceUnavailableException(
                fault_string=fault_string)

        if method != 'DELETE':
            return r.json()

    def post(self, url, args):
        return self.request('POST', url, args)

    def put(self, url, args):
        return self.request('PUT', url, args)

    def delete(self, url):
        self.request('DELETE', url)

    def get(self, url):
        return self.request('GET', url)


class OctaviaDriver(driver_base.LoadBalancerBaseDriver):

    def __init__(self, plugin):
        super(OctaviaDriver, self).__init__(plugin)
        self.req = OctaviaRequest(cfg.CONF.octavia.base_url,
                                  keystone.get_session())

        self.load_balancer = LoadBalancerManager(self)
        self.listener = ListenerManager(self)
        self.pool = PoolManager(self)
        self.member = MemberManager(self)
        self.health_monitor = HealthMonitorManager(self)
        self.l7policy = L7PolicyManager(self)
        self.l7rule = L7RuleManager(self)
        self.octavia_consumer = octavia_messaging_consumer.OctaviaConsumer(
            self)
        service.launch(cfg.CONF, self.octavia_consumer)
        LOG.debug("OctaviaDriver: initialized, version=%s", VERSION)

    @property
    def allocates_vip(self):
        return self.load_balancer.allocates_vip


class LoadBalancerManager(driver_base.BaseLoadBalancerManager):

    @staticmethod
    def _url(lb, id=None):
        s = '/v1/loadbalancers'
        if id:
            s += '/%s' % id
        return s

    @property
    def allows_create_graph(self):
        return True

    @property
    def allows_healthmonitor_thresholds(self):
        return True

    @property
    def allocates_vip(self):
        return cfg.CONF.octavia.allocates_vip

    @property
    def deletes_cascade(self):
        return True

    def _construct_args(self, db_lb, create=True, graph=False):
        args = {'name': db_lb.name,
                'description': db_lb.description,
                'enabled': db_lb.admin_state_up}
        if not create:
            return args

        create_args = {'project_id': db_lb.tenant_id, 'id': db_lb.id,
                       'vip': {'subnet_id': db_lb.vip_subnet_id,
                               'ip_address': db_lb.vip_address,
                               'port_id': db_lb.vip_port_id}}
        args.update(create_args)

        if not graph:
            return args

        if db_lb.listeners:
            args['listeners'] = []
        for db_listener in db_lb.listeners:
            listener_args = self.driver.listener._construct_args(db_listener,
                                                                 graph=True)
            args['listeners'].append(listener_args)
        return args

    def create_and_allocate_vip(self, context, lb):
        self.create(context, lb)

    @async_op
    def create(self, context, lb):
        graph = (lb.listeners and len(lb.listeners) > 0)
        args = self._construct_args(lb, graph=graph)
        self.driver.req.post(self._url(lb), args)

    @async_op
    def update(self, context, old_lb, lb):
        args = self._construct_args(lb, create=False)
        self.driver.req.put(self._url(lb, lb.id), args)

    @async_op
    def delete(self, context, lb):
        self.driver.req.delete(self._url(lb, lb.id))

    @async_op
    def refresh(self, context, lb):
        pass

    def stats(self, context, lb):
        return {}  # todo

    def get(self, lb):
        return self.driver.req.get(self._url(lb, lb.id))

    @async_op
    def delete_cascade(self, context, lb):
        self.driver.req.delete(self._url(lb, lb.id) + '/delete_cascade')


class ListenerManager(driver_base.BaseListenerManager):

    @staticmethod
    def _url(listener, id=None):
        s = '/v1/loadbalancers/%s/listeners' % listener.loadbalancer.id
        if id:
            s += '/%s' % id
        return s

    def _construct_args(self, listener, create=True, graph=False):
        sni_container_ids = [sni.tls_container_id
                             for sni in listener.sni_containers]
        args = {
            'name': listener.name,
            'description': listener.description,
            'enabled': listener.admin_state_up,
            'protocol': listener.protocol,
            'protocol_port': listener.protocol_port,
            'connection_limit': listener.connection_limit,
            'tls_certificate_id': listener.default_tls_container_id,
            'default_pool_id': listener.default_pool_id,
            'sni_containers': sni_container_ids
        }

        if not create:
            return args

        args['project_id'] = listener.tenant_id
        args['id'] = listener.id

        if not graph:
            return args

        del args['default_pool_id']

        if listener.default_pool:
            pool = listener.default_pool
            args['default_pool'] = self.driver.pool._construct_args(pool,
                                                                    graph=True)
        if listener.l7_policies:
            args['l7policies'] = []
            l7_policies = listener.l7_policies
            for l7_policy in l7_policies:
                l7_policy_args = self.driver.l7policy._construct_args(
                    l7_policy, graph=True)
                args['l7policies'].append(l7_policy_args)
        return args

    @async_op
    def create(self, context, listener):
        args = self._construct_args(listener)
        self.driver.req.post(self._url(listener), args)

    @async_op
    def update(self, context, old_listener, listener):
        args = self._construct_args(listener, create=False)
        self.driver.req.put(self._url(listener, id=listener.id), args)

    @async_op
    def delete(self, context, listener):
        self.driver.req.delete(self._url(listener, id=listener.id))


class PoolManager(driver_base.BasePoolManager):

    @staticmethod
    def _url(pool, id=None):
        s = '/v1/loadbalancers/%s/pools' % (
            pool.loadbalancer.id)
        if id:
            s += '/%s' % id
        return s

    def _construct_args(self, pool, create=True, graph=False):
        args = {
            'name': pool.name,
            'description': pool.description,
            'enabled': pool.admin_state_up,
            'protocol': pool.protocol,
            'lb_algorithm': pool.lb_algorithm
        }
        if pool.session_persistence:
            args['session_persistence'] = {
                'type': pool.session_persistence.type,
                'cookie_name': pool.session_persistence.cookie_name,
            }
        else:
            args['session_persistence'] = None

        if not create:
            return args

        args['project_id'] = pool.tenant_id
        args['id'] = pool.id
        if pool.listeners:
            args['listener_id'] = pool.listeners[0].id

        if not graph:
            return args

        if pool.members:
            args['members'] = []
            for member in pool.members:
                member_args = self.driver.member._construct_args(member)
                args['members'].append(member_args)
        if pool.healthmonitor:
            hm_args = self.driver.health_monitor._construct_args(
                pool.healthmonitor)
            args['health_monitor'] = hm_args
        return args

    @async_op
    def create(self, context, pool):
        args = self._construct_args(pool)
        self.driver.req.post(self._url(pool), args)

    @async_op
    def update(self, context, old_pool, pool):
        args = self._construct_args(pool, create=False)
        self.driver.req.put(self._url(pool, id=pool.id), args)

    @async_op
    def delete(self, context, pool):
        self.driver.req.delete(self._url(pool, id=pool.id))


class MemberManager(driver_base.BaseMemberManager):

    @staticmethod
    def _url(member, id=None):
        s = '/v1/loadbalancers/%s/pools/%s/members' % (
            member.pool.loadbalancer.id,
            member.pool.id)
        if id:
            s += '/%s' % id
        return s

    def _construct_args(self, member, create=True):
        args = {
            'enabled': member.admin_state_up,
            'protocol_port': member.protocol_port,
            'weight': member.weight
        }
        if not create:
            return args

        create_args = {
            'id': member.id,
            'ip_address': member.address,
            'subnet_id': member.subnet_id,
            'project_id': member.tenant_id
        }
        args.update(create_args)

        return args

    @async_op
    def create(self, context, member):
        args = self._construct_args(member)
        self.driver.req.post(self._url(member), args)

    @async_op
    def update(self, context, old_member, member):
        args = self._construct_args(member, create=False)
        self.driver.req.put(self._url(member, member.id), args)

    @async_op
    def delete(self, context, member):
        self.driver.req.delete(self._url(member, member.id))


class HealthMonitorManager(driver_base.BaseHealthMonitorManager):

    @staticmethod
    def _url(hm):
        s = '/v1/loadbalancers/%s/pools/%s/healthmonitor' % (
            hm.pool.loadbalancer.id,
            hm.pool.id)
        return s

    def _construct_args(self, hm, create=True):
        args = {
            'type': hm.type,
            'delay': hm.delay,
            'timeout': hm.timeout,
            'rise_threshold': hm.max_retries,
            'fall_threshold': hm.max_retries_down,
            'http_method': hm.http_method,
            'url_path': hm.url_path,
            'expected_codes': hm.expected_codes,
            'enabled': hm.admin_state_up
        }
        if create:
            args['project_id'] = hm.tenant_id
        return args

    @async_op
    def create(self, context, hm):
        args = self._construct_args(hm)
        self.driver.req.post(self._url(hm), args)

    @async_op
    def update(self, context, old_hm, hm):
        args = self._construct_args(hm, create=False)
        self.driver.req.put(self._url(hm), args)

    @async_op
    def delete(self, context, hm):
        self.driver.req.delete(self._url(hm))


class L7PolicyManager(driver_base.BaseL7PolicyManager):

    @staticmethod
    def _url(l7p, id=None):
        s = '/v1/loadbalancers/%s/listeners/%s/l7policies' % (
            l7p.listener.loadbalancer.id,
            l7p.listener.id)
        if id:
            s += '/%s' % id
        return s

    def _construct_args(self, l7p, create=True, graph=False):
        args = {
            'name': l7p.name,
            'description': l7p.description,
            'action': l7p.action,
            'redirect_url': l7p.redirect_url,
            'position': l7p.position,
            'enabled': l7p.admin_state_up
        }
        if args['action'] == constants.L7_POLICY_ACTION_REJECT:
            del args['redirect_url']
        elif args['action'] == constants.L7_POLICY_ACTION_REDIRECT_TO_POOL:
            args['redirect_pool_id'] = l7p.redirect_pool_id
            del args['redirect_url']
        elif args['action'] == constants.L7_POLICY_ACTION_REDIRECT_TO_URL:
            if args.get('redirect_pool_id'):
                del args['redirect_pool_id']
        if not create:
            return args

        args['id'] = l7p.id

        if not graph:
            return args

        if (l7p.redirect_pool and l7p.action ==
                constants.L7_POLICY_ACTION_REDIRECT_TO_POOL):
            del args['redirect_pool_id']
            pool_args = self.driver.pool._construct_args(l7p.redirect_pool,
                                                         graph=True)
            args['redirect_pool'] = pool_args
        if l7p.rules:
            args['l7rules'] = []
            for rule in l7p.rules:
                rule_args = self.driver.l7rule._construct_args(rule)
                args['l7rules'].append(rule_args)
        return args

    @async_op
    def create(self, context, l7p):
        args = self._construct_args(l7p)
        self.driver.req.post(self._url(l7p), args)

    @async_op
    def update(self, context, old_l7p, l7p):
        args = self._construct_args(l7p, create=False)
        self.driver.req.put(self._url(l7p, id=l7p.id), args)

    @async_op
    def delete(self, context, l7p):
        self.driver.req.delete(self._url(l7p, id=l7p.id))


class L7RuleManager(driver_base.BaseL7RuleManager):

    @staticmethod
    def _url(l7r, id=None):
        s = '/v1/loadbalancers/%s/listeners/%s/l7policies/%s/l7rules' % (
            l7r.policy.listener.loadbalancer.id,
            l7r.policy.listener.id,
            l7r.policy.id)
        if id:
            s += '/%s' % id
        return s

    @classmethod
    def _construct_args(cls, l7r, create=True):
        args = {
            'type': l7r.type,
            'compare_type': l7r.compare_type,
            'key': l7r.key,
            'value': l7r.value,
            'invert': l7r.invert
        }
        if create:
            args['id'] = l7r.id
        return args

    @async_op
    def create(self, context, l7r):
        args = self._construct_args(l7r)
        self.driver.req.post(self._url(l7r), args)

    @async_op
    def update(self, context, old_l7r, l7r):
        args = self._construct_args(l7r, create=False)
        self.driver.req.put(self._url(l7r, id=l7r.id), args)

    @async_op
    def delete(self, context, l7r):
        self.driver.req.delete(self._url(l7r, id=l7r.id))



# Copyright 2014 A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron_lib import constants
from oslo_log import log as logging
import six

from neutron_lbaas.db.loadbalancer import models
from neutron_lbaas.services.loadbalancer import constants as lb_const
from neutron_lbaas.services.loadbalancer import data_models


LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class BaseManagerMixin(object):

    def __init__(self, driver):
        self.driver = driver

    @abc.abstractproperty
    def db_delete_method(self):
        pass

    @abc.abstractmethod
    def create(self, context, obj):
        pass

    @abc.abstractmethod
    def update(self, context, obj_old, obj):
        pass

    @abc.abstractmethod
    def delete(self, context, obj):
        pass

    def _successful_completion_lb_graph(self, context, obj):
        listeners = obj.listeners
        obj.listeners = []
        for listener in listeners:
            # need to maintain the link from the child to the load balancer
            listener.loadbalancer = obj
            pool = listener.default_pool
            l7_policies = listener.l7_policies
            if pool:
                pool.listener = listener
                hm = pool.healthmonitor
                if hm:
                    hm.pool = pool
                    self.successful_completion(context, hm)
                for member in pool.members:
                    member.pool = pool
                    self.successful_completion(context, member)
                self.successful_completion(context, pool)
            if l7_policies:
                for l7policy in l7_policies:
                    l7policy.listener = listener
                    l7rules = l7policy.rules
                    for l7rule in l7rules:
                        l7rule.l7policy = l7policy
                        self.successful_completion(context, l7rule)
                    redirect_pool = l7policy.redirect_pool
                    if redirect_pool:
                        redirect_pool.listener = listener
                        rhm = redirect_pool.healthmonitor
                        if rhm:
                            rhm.pool = redirect_pool
                            self.successful_completion(context, rhm)
                        for rmember in redirect_pool.members:
                            rmember.pool = redirect_pool
                            self.successful_completion(context, rmember)
                        self.successful_completion(context, redirect_pool)
                    self.successful_completion(context, l7policy)
            self.successful_completion(context, listener)
        self.successful_completion(context, obj)

    def successful_completion(self, context, obj, delete=False,
                              lb_create=False):
        """
        Sets the provisioning_status of the load balancer and obj to
        ACTIVE.  Should be called last in the implementor's BaseManagerMixin
        methods for successful runs.

        :param context: neutron_lib context
        :param obj: instance of a
                    neutron_lbaas.services.loadbalancer.data_model
        :param delete: set True if being called from a delete method.  Will
                       most likely result in the obj being deleted from the db.
        :param lb_create: set True if this is being called after a successful
                          load balancer create.
        """
        LOG.debug("Starting successful_completion method after a successful "
                  "driver action.")
        if lb_create and obj.listeners:
            self._successful_completion_lb_graph(context, obj)
            return
        obj_sa_cls = data_models.DATA_MODEL_TO_SA_MODEL_MAP[obj.__class__]
        if delete:
            # Check if driver is responsible for vip allocation.  If the driver
            # is responsible, then it is also responsible for cleaning it up.
            # At this point, the VIP should already be cleaned up, so we are
            # just doing neutron lbaas db cleanup.
            if (obj == obj.root_loadbalancer and
                    self.driver.load_balancer.allocates_vip):
                # NOTE(blogan): this is quite dumb to do but it is necessary
                # so that a false negative pep8 error does not get thrown. An
                # "unexpected-keyword-argument" pep8 error occurs bc
                # self.db_delete_method is a @property method that returns a
                # method.
                kwargs = {'delete_vip_port': False}
                self.db_delete_method(context, obj.id, **kwargs)
            else:
                self.db_delete_method(context, obj.id)
        if obj == obj.root_loadbalancer and delete:
            # Load balancer was deleted and no longer exists
            return
        lb_op_status = None
        lb_p_status = constants.ACTIVE
        if obj == obj.root_loadbalancer:
            # only set the status to online if this an operation on the
            # load balancer
            lb_op_status = lb_const.ONLINE

        # Update the load balancer's vip address and vip port id if the driver
        # was responsible for allocating the vip.
        if (self.driver.load_balancer.allocates_vip and lb_create and
                isinstance(obj, data_models.LoadBalancer)):
            self.driver.plugin.db.update_loadbalancer(
                context, obj.id, {'vip_address': obj.vip_address,
                                  'vip_port_id': obj.vip_port_id})
        if delete:
            # We cannot update the status of obj if it was deleted but if the
            # obj is not a load balancer, the root load balancer should be
            # updated
            if not isinstance(obj, data_models.LoadBalancer):
                self.driver.plugin.db.update_status(
                    context, models.LoadBalancer, obj.root_loadbalancer.id,
                    provisioning_status=lb_p_status,
                    operating_status=lb_op_status)
            return
        obj_op_status = lb_const.ONLINE
        if isinstance(obj, data_models.HealthMonitor):
            # Health Monitor does not have an operating status
            obj_op_status = None
        LOG.debug("Updating object of type %s with id of %s to "
                  "provisioning_status = %s, operating_status = %s",
                  obj.__class__, obj.id, constants.ACTIVE, obj_op_status)
        self.driver.plugin.db.update_status(
            context, obj_sa_cls, obj.id,
            provisioning_status=constants.ACTIVE,
            operating_status=obj_op_status)
        if not isinstance(obj, data_models.LoadBalancer):
            # Only update the status of the root_loadbalancer if the previous
            # update was not the root load balancer so we are not updating
            # it twice.
            self.driver.plugin.db.update_status(
                context, models.LoadBalancer, obj.root_loadbalancer.id,
                provisioning_status=lb_p_status,
                operating_status=lb_op_status)

    def failed_completion(self, context, obj):
        """
        Sets the provisioning status of the obj to ERROR.  If obj is a
        loadbalancer it will be set to ERROR, otherwise set to ACTIVE. Should
        be called whenever something goes wrong (raised exception) in an
        implementor's BaseManagerMixin methods.

        :param context: neutron_lib context
        :param obj: instance of a
                    neutron_lbaas.services.loadbalancer.data_model
        """
        LOG.debug("Starting failed_completion method after a failed driver "
                  "action.")
        if isinstance(obj, data_models.LoadBalancer):
            LOG.debug("Updating load balancer %s to provisioning_status = "
                      "%s, operating_status = %s.",
                      obj.root_loadbalancer.id, constants.ERROR,
                      lb_const.OFFLINE)
            self.driver.plugin.db.update_status(
                context, models.LoadBalancer, obj.root_loadbalancer.id,
                provisioning_status=constants.ERROR,
                operating_status=lb_const.OFFLINE)
            return
        obj_sa_cls = data_models.DATA_MODEL_TO_SA_MODEL_MAP[obj.__class__]
        LOG.debug("Updating object of type %s with id of %s to "
                  "provisioning_status = %s, operating_status = %s",
                  obj.__class__, obj.id, constants.ERROR,
                  lb_const.OFFLINE)
        self.driver.plugin.db.update_status(
            context, obj_sa_cls, obj.id,
            provisioning_status=constants.ERROR,
            operating_status=lb_const.OFFLINE)
        LOG.debug("Updating load balancer %s to "
                  "provisioning_status = %s", obj.root_loadbalancer.id,
                  constants.ACTIVE)
        self.driver.plugin.db.update_status(
            context, models.LoadBalancer, obj.root_loadbalancer.id,
            provisioning_status=constants.ACTIVE)

    def update_vip(self, context, loadbalancer_id, vip_address,
                   vip_port_id=None):
        lb_update = {'vip_address': vip_address}
        if vip_port_id:
            lb_update['vip_port_id'] = vip_port_id
        self.driver.plugin.db.update_loadbalancer(context, loadbalancer_id,
                                                  lb_update)


@six.add_metaclass(abc.ABCMeta)
class BaseRefreshMixin(object):

    @abc.abstractmethod
    def refresh(self, context, obj):
        pass


@six.add_metaclass(abc.ABCMeta)
class BaseStatsMixin(object):

    @abc.abstractmethod
    def stats(self, context, obj):
        pass
